#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Start the PyQt graphical interface to gaitutils

@author: Jussi (jnu@iki.fi)
"""

from gaitutils.gui import _gaitmenu

if __name__ == '__main__':
    _gaitmenu.main()



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Stuff related to Python environment

@author: Jussi (jnu@iki.fi)
"""

import sys
import traceback
import subprocess
import os.path as op
from ulstools import env
from pkg_resources import resource_filename
import logging
import hashlib
import os
import tempfile
import binascii

from .gui._windows import error_exit

# Py2: import backported lru_cache
if sys.version_info.major == 2:
    from functools32 import lru_cache
else:
    from functools import lru_cache

logger = logging.getLogger(__name__)


pkg_dir = resource_filename('gaitutils', '')  # package directory
pkg_parent = op.abspath(op.join(pkg_dir, op.pardir))
git_mode = op.isdir(op.join(pkg_parent, '.git'))


class GaitDataError(Exception):
    """Custom exception class to indicate gait data related errors"""

    pass


def _ipython_setup():
    """Performs some IPython magic if we are running in IPython"""
    try:
        __IPYTHON__
    except NameError:
        return
    from IPython import get_ipython

    ip = get_ipython()
    # ip.magic("gui qt5")  # needed for mayavi plots
    # ip.magic("matplotlib qt")  # do mpl plots in separate windows
    ip.magic("reload_ext autoreload")  # these will enable module autoreloading
    ip.magic("autoreload 2")
    # print('warning: setting precision=3 for numpy array printing')
    # np.set_printoptions(precision=3)


def _make_gaitutils_shortcut():
    """Makes a desktop shortcut to gaitmenu gui."""
    env.make_shortcut('gaitutils', 'gui/gaitmenu.py', 'gaitutils menu')


def _git_autoupdate():
    """Update the package git repository.

    This works, if the package was installed into user directory by cloning the
    git repository and running 'python setup.py develop'. In this case, updating
    the cloned repository will effectively update the package.

    The normal way to install the package is via pip install. In this case, the
    package must be updated manually by pip.

    Since this update mechanism is a bit fragile, it is not used by default.

    Return True if update was ran, else False.
    """

    if git_mode:
        logger.debug('running git autoupdate')
        try:
            o = subprocess.check_output(['git', 'pull'], cwd=pkg_parent)
        except subprocess.CalledProcessError:
            logger.warning('git pull returned exit status 1')
            return False
        # check git output to see if update was done; this is fragile
        # better idea might be to use python-git
        updated = 'pdating' in o
        if updated:
            logger.debug('autoupdate status: %s' % o)
            return True
        else:
            logger.debug('package already up to date')
            return False
    else:  # not a git repo
        logger.debug('%s is not a git repo, not running autoupdate' % pkg_parent)
        return False


def _register_gui_exception_handler(full_traceback=False):
    """Registers an exception handler that reports exceptions via GUI"""
    from .config import cfg

    def _my_excepthook(type_, value, tback):
        """Custom exception handler for fatal (unhandled) exceptions:
        report to user via GUI and terminate."""
        # exception and message, but no traceback
        tbackstr = tback if full_traceback else ''
        msg = ''.join(traceback.format_exception(type_, value, tbackstr))
        error_exit(msg)
        # just the message (e.g. ValueError: "blah" -> "blah")
        # may sometimes be confusing, since type of exception is not printed
        # error_exit(value)
        #
        sys.__excepthook__(type_, value, tback)
        sys.exit()

    if cfg.general.gui_exceptions:
        sys.excepthook = _my_excepthook


def lru_cache_checkfile(fun):
    """Cache function results, unless the argument file has changed.

    A lru_cache type decorator for functions that take a file name argument.
    Makes sense for functions that read a file and take a long time to process
    the data (anything that takes significantly longer than the md5 digest).
    Works by computing the md5 digest for the input file and passing that on to
    lru_cache, so the cache is invalidated if file contents have changed.

    Parameters
    ----------
    fun : function
        The function to cache.

    Returns
    -------
    function
        The cached function.
    """

    @lru_cache()
    def cached_fun(filename, md5sum):
        return fun(filename)

    def wrapper(filename):
        with open(filename, 'rb') as f:
            data = f.read()
        md5sum = hashlib.md5(data).hexdigest()
        return cached_fun(filename, md5sum)

    return wrapper


def _named_tempfile(suffix=None):
    """Return a name for a temporary file.
    Does not open the file. Cross-platform. Replaces tempfile.NamedTemporaryFile
    which behaves strangely on Windows.
    """
    LEN = 12  # length of basename
    if suffix is None:
        suffix = ''
    elif suffix[0] != '.':
        raise ValueError('Invalid suffix, must start with dot')
    basename = os.urandom(LEN)  # get random bytes
    # convert to hex string
    # Py2
    if sys.version_info.major == 2:
        basename = binascii.b2a_hex(basename)
    else:
        basename = basename.hex()
    return os.path.join(tempfile.gettempdir(), basename + suffix)



# -*- coding: utf-8 -*-
"""
Reporting in text format.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import absolute_import

import logging
import numpy as np
import os.path as op

from .translations import translate
from .. import utils, sessionutils
from ..config import cfg
from ..viz.plot_common import _compose_varname, _nested_get, _var_unit
from ..timedist import _group_analysis_trials

logger = logging.getLogger(__name__)


# XXX: hardcoded time-distance variables, to set a certain order
_timedist_vars = [
    'Walking Speed',
    'Cadence',
    'Foot Off',
    'Opposite Foot Off',
    'Opposite Foot Contact',
    'Double Support',
    'Single Support',
    'Stride Time',
    'Stride Length',
    'Step Width',
    'Step Length',
]


def _curve_extracted_text(curve_vals, vardefs_dict):
    """Write out curve extracted values as textual tables.

    Works for single and multiple sessions.
    Yields table rows. Use '\n'.join(_curve_extracted_text(curve_vals, vardefs_dict))
    to create the table.
    """
    # we use right/left for textual data (consistency with patient system)
    contexts = utils.get_contexts(right_first=True)
    # write out main title
    is_comparison = len(curve_vals) > 1
    sessions_str = ' vs. '.join(curve_vals.keys())
    yield '\nCurve extracted values for %s' % sessions_str
    # write the pages
    for title, vardefs in vardefs_dict.items():
        COL_WIDTH = 20  # width of the columns containing the data
        # the page title
        yield '\n\n%s' % title
        yield '-' * len(title)
        rowtitle_len = max(len(_compose_varname(vardef)) for vardef in vardefs) + 5
        for session, session_vals in curve_vals.items():
            # session header (only for multiple sessions)
            if is_comparison:
                yield '\n%s:' % session
            # context header
            hdr = ' ' * rowtitle_len
            hdr += ''.join(ctxt_name.ljust(COL_WIDTH) for _, ctxt_name in contexts)
            yield hdr
            for vardef in vardefs:
                # output rows consisting of variable desc followed by R/L values
                row = _compose_varname(vardef).ljust(rowtitle_len)
                for ctxt, _ in contexts:
                    vardef_ctxt = [ctxt + vardef[0]] + vardef[1:]
                    if vardef_ctxt[0] not in session_vals:
                        logger.debug(
                            '%s was not collected for this session' % vardef_ctxt[0]
                        )
                        continue
                    this_vals = _nested_get(session_vals, vardef_ctxt)
                    mean, std = np.mean(this_vals), np.std(this_vals)
                    unit = _var_unit(vardef_ctxt)
                    if unit == 'deg':
                        unit = u'\u00B0'  # Unicode degree sign
                    else:
                        unit = ' ' + unit
                    element = u'%.2f±%.2f%s' % (mean, std, unit)
                    row += element.ljust(COL_WIDTH)
                yield row


def _print_analysis_table(trials):
    """Print analysis vars as text table"""
    res_avg_all, _ = _group_analysis_trials(trials)
    hdr = '%-25s%-9s%-9s\n' % ('Variable', 'Right', 'Left')
    yield hdr
    for _, cond_data in res_avg_all.items():
        for var, val in cond_data.items():
            li = '%-25s%-9.2f%-9.2f%s' % (var, val['Right'], val['Left'], val['unit'])
            yield li


def _print_analysis_text(trials, vars_=None, main_label=None):
    """Print time-distance variables as text"""
    COL_WIDTH = 25  # width of the columns containing the data
    contexts = utils.get_contexts(right_first=True)
    if vars_ is None:
        vars_ = _timedist_vars
    res_avg_all, res_std_all = _group_analysis_trials(trials)
    hdr = '\n%s\n' % translate('Time-distance variables')
    hdr += '-' * len(hdr)
    hdr += '\n\n%s:' % main_label if main_label else '\n\n'
    yield hdr
    for cond, cond_data in res_avg_all.items():
        varnames_trans = [translate(varname) for varname in vars_]
        rowtitle_len = max(len(varname) for varname in varnames_trans) + 5
        hdr = ' ' * rowtitle_len
        hdr += ''.join(
            translate(ctxt_name).ljust(COL_WIDTH) for _, ctxt_name in contexts
        )
        yield hdr
        for varname, varname_trans in zip(vars_, varnames_trans):
            val = cond_data[varname]
            val_std = res_std_all[cond][varname]
            unit = translate(val['unit'])
            li = u''
            li += varname_trans.ljust(rowtitle_len)
            for _, ctxt_name in contexts:
                element = u'%.2f±%.2f %s' % (val[ctxt_name], val_std[ctxt_name], unit)
                li += element.ljust(COL_WIDTH)
            yield li
    yield ''


def _session_analysis_text(sessionpath):
    """Return session time-distance vars as text"""

    sessiondir = op.split(sessionpath)[-1]
    tagged_trials = sessionutils.get_c3ds(
        sessionpath, tags=cfg.eclipse.tags, trial_type='dynamic'
    )
    return '\n'.join(
        _print_analysis_text({sessiondir: tagged_trials}, main_label=sessiondir)
    )



# -*- coding: utf-8 -*-
"""
Translate text.

@author: Jussi (jnu@iki.fi)
"""

import logging

from ..config import cfg

logger = logging.getLogger(__name__)


translations = dict()
translations['finnish'] = {
    'Comparison report': u'Vertailuraportti',
    'Patient code': u'Potilaskoodi',
    'Session date': u'Mittauksen päivämäärä',
    'Normal data': u'Normaalidata',
    'Session': u'Mittaus',
    'Description': u'Kuvaus',
    'vuotta': u'years',
    'Age at time of measurement': u'Ikä mittaushetkellä',
    'unknown': u'ei tiedossa',
    'Social security number': u'Henkilötunnus',
    'Name': u'Nimi',
    'Results of gait analysis': u'Kävelyanalyysin tulokset',
    'Right': u'Oikea',
    'Left': u'Vasen',
    'R': u'O',
    'L': u'V',
    'Time-distance variables': u'Matka-aikamuuttujat',
    'Single Support': u'Yksöistukivaihe',
    'Double Support': u'Kaksoistukivaihe',
    'Opposite Foot Contact': u'Vastakkaisen jalan kontakti',
    'Opposite Foot Off': u'Vastakkainen jalka irti',
    'Limp Index': u'Limp-indeksi',
    'Step Length': u'Askelpituus',
    'Foot Off': u'Tukivaiheen kesto',
    'Walking Speed': u'Kävelynopeus',
    'Stride Length': u'Askelsyklin pituus',
    'Step Width': u'Askelleveys',
    'Step Time': u'Askeleen kesto',
    'Cadence': u'Kadenssi',
    'Stride Time': u'Askelsyklin kesto',
    'steps/min': u'1/min',
}

# rewrite keys in lower case
translations = {key.lower(): val for key, val in translations.items()}
for key, val in translations.items():
    translations[key] = {lang.lower(): trans_di for lang, trans_di in val.items()}


def translate(text):
    """Simple mechanism for translating words and expressions.

    The translation language is taken from the config. Currently supported
    values: 'finnish'. If None, do not translate.

    Parameters
    ----------
    text : str
        Word or expression to translate.

    Returns
    -------
    str
        The translated text.
    """
    global translations
    language = cfg.report.language
    if language is None:
        return text
    elif language not in translations:
        raise ValueError('Unknown translation language')
    elif text.lower() not in translations[language]:
        logger.info('no translation for %s' % text)
        return text
    else:
        return translations[language][text.lower()]



# -*- coding: utf-8 -*-
"""
PyQt dialogs etc.

@author: Jussi (jnu@iki.fi)
"""

from PyQt5 import uic, QtWidgets
from PyQt5.QtWidgets import QDialogButtonBox
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from pkg_resources import resource_filename
import os.path as op
import ast
import io
from collections import defaultdict
import logging

from .. import nexus
from ..envutils import GaitDataError
import configdot
from ..config import _handle_cfg_defaults, cfg_user_fn, cfg

logger = logging.getLogger(__name__)


def qt_matplotlib_window(fig):
    """Show matplotlib figure fig in new Qt window. Window is returned"""
    _mpl_win = QtWidgets.QDialog()
    # _mpl_win.setGeometry(100, 100, 1500, 1000)
    _mpl_win._canvas = FigureCanvas(fig)
    _mpl_win._canvas.setSizePolicy(
        QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
    )
    _mpl_win._canvas.updateGeometry()  # not sure if this does anything
    toolbar = NavigationToolbar(_mpl_win._canvas, _mpl_win)
    layout = QtWidgets.QVBoxLayout()
    layout.addWidget(toolbar)
    layout.addWidget(_mpl_win._canvas)
    layout.setSpacing(0)
    _mpl_win.setLayout(layout)
    _mpl_win._canvas.draw()
    _mpl_win.show()
    return _mpl_win


def qt_message_dialog(msg):
    """Show message with 'OK' button"""
    dlg = QtWidgets.QMessageBox()
    dlg.setWindowTitle('Message')
    dlg.setText(msg)
    dlg.addButton(QtWidgets.QPushButton('Ok'), QtWidgets.QMessageBox.YesRole)
    dlg.exec_()


def qt_yesno_dialog(msg):
    """Show message with Yes and No buttons, return role accordingly"""
    dlg = QtWidgets.QMessageBox()
    dlg.setWindowTitle('Confirm')
    dlg.setText(msg)
    dlg.addButton(QtWidgets.QPushButton('Yes'), QtWidgets.QMessageBox.YesRole)
    dlg.addButton(QtWidgets.QPushButton('No'), QtWidgets.QMessageBox.NoRole)
    dlg.exec_()
    return dlg.buttonRole(dlg.clickedButton())


def qt_dir_chooser():
    """Selector dialog to select dir (or multiple dirs). Always returns a list
    (empty if dialog was canceled)"""
    # native dialog - single dir only
    dir = QtWidgets.QFileDialog.getExistingDirectory(None, 'Select session')
    return [dir] if dir else list()
    # non-native dialog - multiple dirs. a bit messy, currently not in use
    file_dialog = QtWidgets.QFileDialog()
    file_dialog.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
    file_dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, True)
    file_view = file_dialog.findChild(QtWidgets.QListView, 'listView')
    if file_view:
        file_view.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
    f_tree_view = file_dialog.findChild(QtWidgets.QTreeView)
    if f_tree_view:
        f_tree_view.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
    return file_dialog.selectedFiles() if file_dialog.exec_() else []


class QCompoundEditorDict(QtWidgets.QDialog):
    """Compound editor for dicts"""

    def __init__(self, data, parent=None):
        super(QCompoundEditorDict, self).__init__(parent=parent)
        root_layout = QtWidgets.QVBoxLayout()
        self.datable = QtWidgets.QTableWidget()
        self.data = data
        # create the button box
        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        insertbutton = QtWidgets.QPushButton('Insert')
        deletebutton = QtWidgets.QPushButton('Delete')
        self.buttonbox.addButton(insertbutton, QDialogButtonBox.ActionRole)
        self.buttonbox.addButton(deletebutton, QDialogButtonBox.ActionRole)
        insertbutton.clicked.connect(self._insert_item)
        deletebutton.clicked.connect(self._delete_item)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        self.datable = QtWidgets.QTableWidget()
        self._init_table()
        root_layout.addWidget(self.datable)
        root_layout.addWidget(self.buttonbox)
        self.datable.resizeColumnsToContents()
        self.datable.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.setLayout(root_layout)
        self.setWindowTitle('Edit config item')
        if not data:
            self._insert_item()

    def _init_table(self):
        """Init table for dict data"""
        self.datable.setColumnCount(2)
        self.datable.setHorizontalHeaderLabels(['Key', 'Value'])
        for k, (key, val) in enumerate(self.data.items()):
            key_txt, val_txt = repr(key), repr(val)
            self.datable.insertRow(k)
            self.datable.setItem(k, 0, QtWidgets.QTableWidgetItem(key_txt))
            self.datable.setItem(k, 1, QtWidgets.QTableWidgetItem(val_txt))

    def _insert_item(self):
        """Insert a new dict/list item at current position and start editing it"""
        pos = self.datable.currentRow()
        self.datable.insertRow(pos + 1)
        newitem = QtWidgets.QTableWidgetItem()
        self.datable.setItem(pos + 1, 0, newitem)
        self.datable.editItem(newitem)

    def _delete_item(self):
        """Delete dict/list item at current position"""
        pos = self.datable.currentRow()
        self.datable.removeRow(pos)

    def _collect_column(self, col):
        """Collect data from column col"""
        items = (self.datable.item(row, col) for row in range(self.datable.rowCount()))
        items = (it for it in items if it is not None)
        return (it.text() for it in items)

    def accept(self):
        """Do some sanity checks, close dialog if ok"""
        keys, vals = self._collect_column(0), self._collect_column(1)
        _text_data = {k: v for k, v in zip(keys, vals)}
        # keys and vals are now strings (hopefully) representing valid Python expressions
        # return a dict with the actual evaluated expressions
        _result = dict()
        for row, (key_txt, val_txt) in enumerate(_text_data.items(), 1):
            try:
                key = ast.literal_eval(key_txt)
                _result[key] = ast.literal_eval(val_txt)
            except (SyntaxError, ValueError):
                qt_message_dialog(
                    'Invalid input for item %s: %s on row %d\n'
                    'Inputs must be valid Python expressions (e.g. strings must be quoted).\n'
                    'Please fix before closing or cancel dialog'
                    % (key_txt, val_txt, row)
                )
                return
        self.data = _result
        self.done(QtWidgets.QDialog.Accepted)


class QCompoundEditorList(QCompoundEditorDict):
    """Compound editor for lists"""

    def _init_table(self):
        """Init table for list data"""
        self.datable.setColumnCount(1)
        self.datable.setHorizontalHeaderLabels(['Value'])
        for k, val in enumerate(self.data):
            val_txt = repr(val)
            self.datable.insertRow(k)
            self.datable.setItem(k, 0, QtWidgets.QTableWidgetItem(val_txt))

    def accept(self):
        """Do some sanity checks, close dialog if ok"""
        # list elements are now strings (hopefully) representing valid Python expressions
        # return a list with the actual evaluated expressions
        _text_data = list(self._collect_column(0))
        _result = list()
        for row, item_txt in enumerate(_text_data, 1):
            try:
                item = ast.literal_eval(item_txt)
                _result.append(item)
            except (SyntaxError, ValueError):
                qt_message_dialog(
                    'Invalid input for item %s on row %d.\n'
                    'Inputs must be valid Python expressions (e.g. strings must be quoted).\n'
                    'Please fix before closing or cancel dialog' % (item_txt, row)
                )
                return
        self.data = _result
        self.done(QtWidgets.QDialog.Accepted)


class OptionsDialog(QtWidgets.QDialog):
    """Dialog for changing gaitutils options"""

    def _create_tab(self, section, secname):
        """Create a tab for the tab widget, according to config items"""
        tab = QtWidgets.QWidget()
        lout = QtWidgets.QFormLayout()
        tab.setLayout(lout)
        # get items sorted by comment
        items = sorted(
            (item for (itname, item) in section),
            key=lambda it: configdot.get_description(it),
        )
        # loop through config items and insert appropriate input widget for each
        for item in items:
            _save_widget = True
            desc = configdot.get_description(item)
            if not desc:
                raise RuntimeError('Config item %s missing a description' % item)
            if isinstance(item.value, bool):  # use simple checkbox for boolean items
                input_widget = QtWidgets.QCheckBox()
                input_widget.setChecked(item.value)
            elif isinstance(item.value, list) or isinstance(item.value, dict):
                # launch the compound editor for complex types (list and dict)
                input_widget = QtWidgets.QPushButton()
                input_widget.setText('Edit...')
                # lambda needs to consume the extra value from connect(), hence x
                input_widget.clicked.connect(
                    lambda x, it=item: self._edit_with_compound_editor(it)
                )
                # do not register compound editor buttons as cfg widgets, since the compound editor
                # callback updates cfg by itsel
                _save_widget = False
            else:
                # for any other types, use a line edit with the literal value
                input_widget = QtWidgets.QLineEdit()
                input_widget.setText(item.literal_value)
                input_widget.setCursorPosition(0)  # show beginning of line
            lout.addRow(desc, input_widget)
            if _save_widget:
                self._input_widgets[secname][item.name] = input_widget
        return tab

    def __init__(self, parent, default_tab=0):
        super(self.__class__, self).__init__(parent)
        _main_layout = QtWidgets.QVBoxLayout(self)
        self._input_widgets = defaultdict(lambda: dict())

        # build button box
        std_buttons = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(std_buttons)
        loadButton = QtWidgets.QPushButton('Load...')
        saveButton = QtWidgets.QPushButton('Save...')
        self.buttonBox.addButton(loadButton, QtWidgets.QDialogButtonBox.ActionRole)
        self.buttonBox.addButton(saveButton, QtWidgets.QDialogButtonBox.ActionRole)
        loadButton.clicked.connect(self.load_config_dialog)
        saveButton.clicked.connect(self.save_config_dialog)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        # build tabs according to cfg
        self.tabWidget = QtWidgets.QTabWidget()
        secs = sorted(((secname, sec) for secname, sec in cfg), key=lambda tup: tup[0])
        for secname, sec in secs:
            desc = configdot.get_description(sec) or secname
            tab = self._create_tab(sec, secname)
            self.tabWidget.addTab(tab, desc)

        _main_layout.addWidget(self.tabWidget)
        helptext = QtWidgets.QLabel()
        helptext.setText(
            'Changes into options will stay in effect until the program is restarted. To make changes\n'
            'permanent and automatically loaded on startup, save them into\n%s'
            % cfg_user_fn
        )
        _main_layout.addWidget(helptext)
        _main_layout.addWidget(self.buttonBox)
        self.setLayout(_main_layout)
        self.setWindowTitle('Edit configuration')

    def _edit_with_compound_editor(self, item):
        """Opens a compound editor for selected config item"""
        val = item.value
        editor = QCompoundEditorDict if isinstance(val, dict) else QCompoundEditorList
        dlg = editor(val)
        if dlg.exec_():
            item.value = dlg.data

    def load_config_dialog(self):
        """Bring up load dialog and load selected file"""
        fout = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Load config file', op.expanduser('~'), 'Config files (*.cfg)'
        )
        fname = fout[0]
        if fname:
            try:
                cfg_new = configdot.parse_config(fname)
                configdot.update_config(
                    cfg,
                    cfg_new,
                    create_new_sections=False,
                    create_new_items=['layouts'],
                    update_comments=False,
                )
            except ValueError:
                qt_message_dialog('Could not parse %s' % fname)
            else:
                self._update_inputs()

    def save_config_dialog(self):
        """Bring up save dialog and save data"""
        wname, txt = self._update_cfg()
        if wname is not None:
            qt_message_dialog(
                'Invalid input for item %s: %s\n'
                'Please fix before saving' % (wname, txt)
            )
        else:
            fout = QtWidgets.QFileDialog.getSaveFileName(
                self, 'Save config file', op.expanduser('~'), 'Config files (*.cfg)'
            )
            fname = fout[0]
            if fname:
                with io.open(fname, 'w', encoding='utf8') as f:
                    txt = configdot.dump_config(cfg)
                    f.writelines(txt)

    def _update_inputs(self):
        """Update value-visible input widgets according to current cfg"""
        for secname, sec in cfg:
            for itemname, item in sec:
                if itemname not in self._input_widgets[secname]:
                    continue
                _widget = self._input_widgets[secname][itemname]
                val = item.literal_value
                if isinstance(_widget, QtWidgets.QLineEdit):
                    _widget.setText(val)
                    _widget.setCursorPosition(0)
                elif isinstance(_widget, QtWidgets.QCheckBox):
                    _widget.setChecked(item.value)

    def _update_cfg(self):
        """Update cfg according to input widgets"""
        for secname, sec in cfg:
            for itemname, item in sec:
                if itemname not in self._input_widgets[secname]:
                    continue
                _widget = self._input_widgets[secname][itemname]
                if isinstance(_widget, QtWidgets.QLineEdit):
                    try:
                        item.value = ast.literal_eval(_widget.text())
                    except (SyntaxError, ValueError):
                        return itemname, _widget.text()
                elif isinstance(_widget, QtWidgets.QCheckBox):
                    item.value = _widget.isChecked()
                else:
                    raise RuntimeError('Invalid input widget class, how come?')
        _handle_cfg_defaults(cfg)
        return None, None

    def accept(self):
        """Update config and close dialog, if widget inputs are ok. Otherwise show error dialog"""
        wname, txt = self._update_cfg()
        if wname is not None:
            qt_message_dialog(
                'Invalid input for item %s: %s\n'
                'Please fix before closing or cancel dialog' % (wname, txt)
            )
        else:
            self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept


class ChooseSessionsDialog(QtWidgets.QDialog):
    """A dialog for picking report sessions"""

    def __init__(self, min_sessions=1, max_sessions=3):
        QtWidgets.QDialog.__init__(self)
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/sessions.ui')
        uic.loadUi(uifile, self)
        # self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.btnBrowseSession.clicked.connect(self.add_session)
        self.btnAddNexusSession.clicked.connect(
            lambda: self.add_session(from_nexus=True)
        )
        self.btnClearAll.clicked.connect(self.listSessions.clear)
        self.btnClearCurrent.clicked.connect(self.listSessions.rm_current_item)
        self.max_sessions = max_sessions
        self.min_sessions = min_sessions

    def add_session(self, from_nexus=False):
        if len(self.sessions) == self.max_sessions:
            qt_message_dialog(
                'You can specify maximum of %d sessions' % self.max_sessions
            )
            return
        if from_nexus:
            try:
                dirs = [nexus.get_sessionpath()]
            except GaitDataError:
                qt_message_dialog('Cannot get session path from Nexus')
                return
        else:
            dirs = qt_dir_chooser()
        dirs = [op.normpath(d) for d in dirs]
        for dir_ in dirs:
            if dir_ in self.sessions:
                qt_message_dialog('Session %s already loaded' % dir_)
            else:
                self.listSessions.add_item(dir_, data=dir_)

    @property
    def sessions(self):
        return [item.userdata for item in self.listSessions.items]

    def accept(self):
        if len(self.sessions) < self.min_sessions:
            qt_message_dialog(
                'Please select at least %d session%s'
                % (self.min_sessions, 's' if self.min_sessions > 1 else '')
            )
        else:
            self.done(QtWidgets.QDialog.Accepted)



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 23 10:59:28 2016

Eclipse (database) hacks.

@author: Jussi (jnu@iki.fi)
"""
from builtins import str
from builtins import object
import logging
import io
import configobj
from configobj import ConfigObj
from collections import defaultdict, OrderedDict

from .numutils import _isint
from .envutils import GaitDataError

logger = logging.getLogger(__name__)


class FileFilter(object):
    """Filter class for configobj.

    File-like class that will replace strings according to the replace
    dict below. This is needed for prefiltering before configobj parsing,
    since configobj will not tolerate lines with neither key nor value
    (which seem to occasionally appear in Eclipse files).
    Also deduplicates successive identical lines for same reason.
    """

    replace = {'\n=\n': '\n'}

    def __init__(self, fname):
        self.fname = fname

    def read(self):
        """Read data.

        ConfigObj seems to use only this method.
        """
        # Eclipse switched from latin-1 to utf-8 at some point...
        try:
            self.fp = io.open(self.fname, encoding='utf8')
            data = self.fp.read()
        except UnicodeDecodeError:
            logger.warning('Cannot interpret %s as utf-8, trying latin-1' % self.fname)
            self.fp = io.open(self.fname, encoding='latin-1')
            data = self.fp.read()
        # filter
        for val, newval in FileFilter.replace.items():
            data = data.replace(val, newval)
        # rm subsequent duplicate lines - a bit cryptic
        data = '\n'.join(list(OrderedDict.fromkeys(data.split('\n'))))
        return data

    def close(self):
        self.fp.close()


def _enf_reader(fname_enf):
    """Return enf reader (ConfigObj instance)."""
    fp = FileFilter(fname_enf)
    # do not listify comma-separated values
    # logger.debug('loading %s' % fname_enf)
    try:
        cp = ConfigObj(fp, encoding='utf8', list_values=False, write_empty_values=True)
    except configobj.ParseError:
        raise GaitDataError('Cannot parse config file %s' % fname_enf)
    if 'TRIAL_INFO' not in cp.sections:
        raise GaitDataError('No trial info in .enf file')
    return cp


def get_eclipse_keys(fname_enf, return_empty=False):
    """Read key/value pairs from enf file.

    Currently, only keys in the TRIAL_INFO section will be read.

    Parameters
    ----------
    fname_enf : str
        The filename.
    return_empty : bool, optional
        If True, return also keys without value

    Returns
    -------
    dict
        Dict of the eclipse keys and values.
    """
    di = defaultdict(lambda: u'')
    cp = _enf_reader(fname_enf)
    di.update(
        {key: val for key, val in cp['TRIAL_INFO'].items() if val != '' or return_empty}
    )
    return di


def _eclipse_forceplate_keys(eclipse_keys):
    """Filter that returns Eclipse forceplate keys/values as a dict."""
    return {
        key: val
        for key, val in eclipse_keys.items()
        if key[:2] == 'FP' and len(key) == 3 and _isint(key[2])
    }


def set_eclipse_keys(fname_enf, eclipse_dict, update_existing=False):
    """Set key/value pairs in enf file.

    Keys will be written into the TRIAL_INFO section.

    Parameters
    ----------
    fname_enf : str
        The filename.
    eclipse_dict : dict
        A dict containing the new key/value pairs.
    update_existing : bool, optional
        If True, overwrite existing keys.
    """
    cp = _enf_reader(fname_enf)
    did_set = False
    for key, val in eclipse_dict.items():
        if key not in cp['TRIAL_INFO'].keys() or update_existing:
            cp['TRIAL_INFO'][key] = val
            did_set = True
    if did_set:
        logger.debug('writing %s' % fname_enf)
        # output the config lines; this writes utf8-encoded bytes
        out = cp.write()
        # they must be converted to str for writing
        outu = [str(line, encoding='utf8') + '\n' for line in out]
        with io.open(fname_enf, 'w', encoding='utf8') as fp:
            fp.writelines(outu)
    else:
        logger.debug('did not set any keys')



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Class for handling EMG data.

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division
from past.builtins import basestring
from builtins import object
import numpy as np
import logging

from . import read_data, numutils
from .config import cfg

logger = logging.getLogger(__name__)


class EMG(object):
    """Class for processing and storing EMG data.

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.
    correction_factor : int, optional
        After read, the EMG data is multiplied by this factor.
    """

    def __init__(self, source, correction_factor=1):
        logger.debug('new EMG instance from %s' % source)
        self.source = source
        self.passband = cfg.emg.passband
        self._data = None
        self.t = None
        self.sfrate = None
        self.correction_factor = correction_factor

    @property
    def data(self):
        """Get the EMG data.

        The data is cached in an instance variable, so it's read only once
        during the lifetime of the instance.

        Returns
        -------
        dict
            The EMG data, keyed by channel name. Values are shape (N,) ndarrays
            of sample values.
        """
        if self._data is None:
            self._read_data()
        return self._data

    def _read_data(self):
        """Actually read the EMG data from source"""
        meta = read_data.get_metadata(self.source)
        logger.debug('reading EMG from %s' % meta['trialname'])
        self.sfrate = meta['analograte']
        emgdi = read_data.get_emg_data(self.source)
        self._data = emgdi['data']
        self.t = emgdi['t']

    def _edf_export(self, filename):
        """Export the EMG data to EDF format.

        Dumps raw (unfiltered) data from all physical channels to the given EDF
        file. Format is currently set to EDF plus. Uses pyedflib as a soft dependency.

        Parameters
        ----------
        filename : str
            Name of EDF file to write.

        """
        try:
            import pyedflib
        except ImportError:
            raise RuntimeError(
                'You need to install the pyedflib package to use this function'
            )
        # default to EDF+ for the time being
        file_type = pyedflib.FILETYPE_EDFPLUS
        f = pyedflib.EdfWriter(
            filename,
            len(self.data),
            file_type=file_type,
        )
        channel_info = list()
        data_list = list()
        for chname, chdata in self.data.items():
            # scale signals to more typical EMG unit of mV
            chdata_scaled = chdata * 1e3
            # strip Voltage. prefix that Nexus inserts
            if chname.find('Voltage.') == 0:
                chname = chname[8:]
            # try to conform with the EDF channel label standard
            chname = 'EMG %s' % chname
            # EDF stores data as 16 bit ints
            # since ranges can be set per-channel, we map individual
            # range of each signal into the 16-bit digital scale
            nbits = 16
            ch_dict = {
                'label': chname,
                'dimension': 'mV',
                'sample_rate': self.sfrate,
                'physical_max': max(chdata_scaled),
                'physical_min': min(chdata_scaled),
                'digital_max': 2 ** (nbits - 1) - 1,
                'digital_min': -(2 ** (nbits - 1)),
                'transducer': 'EMG',
                'prefilter': '',
            }
            channel_info.append(ch_dict)
            data_list.append(chdata_scaled)
        f.setSignalHeaders(channel_info)
        f.writeSamples(data_list)
        f.close()

    def _match_name(self, chname):
        """Fuzzily match channel name"""
        if not (isinstance(chname, basestring) and len(chname)) >= 2:
            raise ValueError('invalid channel name: %s' % chname)
        matches = [x for x in self.data if x.find(chname) >= 0]
        if len(matches) == 0:
            raise KeyError('No matching channel for %s' % chname)
        else:
            ch = min(matches, key=len)  # choose shortest matching name
        if len(matches) > 1:
            logger.warning(
                'multiple channel matches for %s: %s -> %s' % (chname, matches, ch)
            )
        return ch

    def get_channel_data(self, chname, rms=False):
        """Return EMG data for a given channel.

        Uses name matching. The given chname is matched (case sensitive) against
        the channel names in the data source and the shortest match is returned.
        For example, with chname=='LGas', 'Voltage.LGas8' and
        'Voltage.LGas8_filtered' would be matches, and the former would be
        returned.

        Data is returned filtered if self.passband is set.

        Parameters
        ----------
        ch : string
            The EMG channel name. Name matching is used (see above).
        rms : bool
            Return moving-window RMS instead of raw data.

        Returns
        -------
        ndarray
            The data, shape (N,).
        """
        ch = self._match_name(chname)
        data = self.data[ch]
        if rms:
            data = numutils.rms(data, cfg.emg.rms_win)
        elif self.passband:  # no filtering for RMS data
            # filtered data is currently not cached; _filtfilt() call typically
            # takes less than 1 ms, so this should not be a problem, unless a
            # huge number of calls are made
            data = numutils._filtfilt(data, self.passband, self.sfrate)
        data *= self.correction_factor
        return data

    def has_channel(self, chname):
        """Check whether a channel exists in the data.

        Parameters
        ----------
        chname : str
            The desired channel name. Name matching is used (see docstring for
            get_channel_data)

        Returns
        -------
        bool
            True if the channel exists, False otherwise.
        """
        try:
            self._match_name(chname)
        except KeyError:
            return False
        return True

    def status_ok(self, chname):
        """Check whether a channel exists and has valid EMG signal.

        Parameters
        ----------
        chname : str
            The desired channel name. Name matching is used (see docstring for
            get_channel_data)

        Returns
        -------
        bool
            True if the channel exists and has valid data, False otherwise.
        """
        if not self.has_channel(chname):
            return False
        elif cfg.emg.chs_disabled and chname in cfg.emg.chs_disabled:
            return False
        data = self.get_channel_data(chname)
        return self._is_valid_emg(data) if cfg.emg.autodetect_bads else True

    @staticmethod
    def context_ok(chname, context):
        """Check if the channel context matches given context.

        Parameters
        ----------
        chname : str
            A configured channel name. (Name matching is not applied here.)
        context : str
            Context ('L' or 'R').

        Returns
        -------
        bool
            False if context does not match. True if the context matches or is unknown.
        """
        if (
            chname in cfg.emg.channel_context
            and context.upper() != cfg.emg.channel_context[chname].upper()
        ):
            return False
        return True

    def _is_valid_emg(self, data):
        """Check whether channel contains a valid EMG signal."""
        return cfg.emg.variance_ok[0] < np.var(data) < cfg.emg.variance_ok[1]


class AvgEMG(EMG):
    """Class for storing averaged RMS EMG.

    This tries to match the API of the EMG class, but differs in following ways:
    -precomputed RMS data is stored in self._data
    -only the RMS data can be returned
    -no filtering is done

    Parameters
    ----------
    data : dict
        The averaged data, e.g. from stats.average_analog.data().
    """

    def __init__(self, data):
        self._data = data

    def get_channel_data(self, chname, rms=None):
        if not rms:
            raise RuntimeError('AvgEMG can only return averaged RMS data')
        chname = self._match_name(chname)
        return self._data[chname]

    def status_ok(self, chname):
        return self.has_channel(chname)

    def _is_valid_emg(self, data):
        raise RuntimeError('signal check not implemented for averaged EMG')

    def filt(self, y, passband):
        raise RuntimeError('filtering not implemented for averaged EMG')

    def read(self):
        raise RuntimeError('read not implemented for averaged EMG')



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 14 10:47:20 2019

@author: vicon123
"""

import logging

from gaitutils import autoprocess


logging.basicConfig(level=logging.DEBUG)

autoprocess._do_autoproc(
    u"D:\\ViconData\\Clinical\\D0050_LR\\2019_9_5_postOp2v_itsenäinen_LR\\2019_9_5_postOp2v_itsenäinen_LR04.Trial.enf"
)



# -*- coding: utf-8 -*-
"""
Normal data readers

@author: Jussi (jnu@iki.fi)
"""


from builtins import next
from builtins import zip
import numpy as np
import openpyxl
import os.path as op
import json
import io
import logging

from .config import cfg
from .envutils import GaitDataError
from . import sessionutils, numutils
from .numutils import _isfloat
from ulstools.num import age_from_hetu
from .models import models_all
from .envutils import lru_cache_checkfile


logger = logging.getLogger(__name__)


def _read_configured_model_normaldata(age=None):
    """Read all model normal data defined in config.

    If age is specified, include also age specific normal data, which will take
    preference over other data.
    """
    model_normaldata = dict()
    # we generously accept both list and string
    if isinstance(cfg.general.normaldata_files, list):
        normaldata_files = cfg.general.normaldata_files
    else:
        normaldata_files = [cfg.general.normaldata_files]
    for fn in normaldata_files:
        ndata = _read_model_normaldata_file(fn)
        model_normaldata.update(ndata)
    if age is not None:
        age_ndata_file = _find_normaldata_for_age(age)
        if age_ndata_file:
            age_ndata = _read_model_normaldata_file(age_ndata_file)
            model_normaldata.update(age_ndata)
    return model_normaldata


def _read_configured_emg_normaldata():
    """Read the EMG normal data defined in config."""
    return _read_emg_normaldata_file(cfg.emg.normaldata_file)


def _read_session_normaldata(session):
    """Read model normal data according to patient in given session.

    This is a convenience that figures out the age of the patient
    and calls _read_default_normaldata"""
    info = sessionutils.load_info(session)
    if info is not None and 'hetu' in info:
        age = age_from_hetu(info['hetu'])
    else:
        age = None
    return _read_configured_model_normaldata(age)


@lru_cache_checkfile
def _read_emg_normaldata_file(filename):
    """Read JSON formatted EMG normal data.

    The normaldata is stored as a dict, where keys correspond to electrode names
    and values are 101-element lists (returned as numpy arrays). The values
    correspond to EMG activation in the range 0..1 and the index corresponds to
    % of the gait cycle.
    """
    if filename is None:
        filename = cfg.emg.normaldata_file
    logger.debug('reading EMG normal data from %s' % filename)
    if not op.isfile(filename):
        raise GaitDataError('No such file %s' % filename)
    with io.open(filename, 'r', encoding='utf-8') as f:
        emg_normals = json.load(f)
    for k, v in emg_normals.items():
        if len(v) != 101:
            raise GaitDataError('EMG normal data has invalid dims')
    # convert lists into numpy arrays
    return {k: np.array(v) for k, v in emg_normals.items()}


@lru_cache_checkfile
def _read_model_normaldata_file(filename):
    """Read model normaldata from a file.

    Returns a dict, where keys are variables and values are ndarrays of shape
    (n,2). n is either 1 (for scalar variables) or 51 (data on 0..100% gait
    cycle, defined every 2% of cycle). The first and second columns are min and
    max values, respectively. (Typically mean-stddev and mean+stddev).
    GCD and XLSX (Polygon) formats are currently supported.
    """
    logger.debug('reading normal data from %s' % filename)
    if not op.isfile(filename):
        raise GaitDataError('No such file %s' % filename)
    type_ = op.splitext(filename)[1].lower()
    if type_ == '.gcd':
        return _read_gcd(filename)
    elif type_ == '.xlsx':
        return _read_xlsx(filename)
    else:
        raise GaitDataError('Only .gcd or .xlsx file formats are supported')


def _find_normaldata_for_age(age):
    """Find age specific normaldata file"""
    if not age:
        return None
    for age_range, filename in cfg.general.normaldata_age.items():
        if age_range[0] <= age <= age_range[1]:
            logger.debug('found normal data file %s for age %d' % (filename, age))
            return filename


def _check_normaldata(ndata):
    """Sanity checks for model normaldata"""
    for val in ndata.values():
        if not all(np.diff(val) >= 0):
            raise ValueError('Normal data not in min/max format')
        if val.shape[0] not in [1, 51]:  # must be gait cycle data or scalar
            raise ValueError('Normal data has unexpected dimensions')
    return ndata


def _read_gcd(filename):
    """Read normaldata from a GCD file.

    gcd data is assumed to be in (mean, dev) 2-column format and will converted
    to (min, max) (Polygon normal data format) as (mean-dev, mean+dev). GCD
    variable names are different from models.py and will be translated according
    to each models translation table.
    """
    ndata = dict()
    with open(filename, 'r') as f:
        lines = f.readlines()
    varname = None
    for li in lines:
        lis = li.split()
        if li[0] == '!':  # new variable
            varname = lis[0][1:]
            ndata[varname] = list()
        elif varname and _isfloat(lis[0]):  # actual data
            # assume mean, dev format
            mean, dev = np.array(lis, dtype=float)
            ndata[varname].append([mean - dev, mean + dev])
        else:  # comment etc.
            continue
    # translate variable names
    ndata_ = dict()
    for nvarname, nval in ndata.items():
        for model in models_all:
            if nvarname in model.gcd_normaldata_map:
                logger.debug(
                    'mapping normal data variable %s -> %s'
                    % (nvarname, model.gcd_normaldata_map[nvarname])
                )
                nvarname = model.gcd_normaldata_map[nvarname]
                break
        ndata_[nvarname] = nval
    normaldata = {key: np.array(val) for key, val in ndata_.items()}
    return _check_normaldata(normaldata)


def _read_xlsx(filename):
    """Read normal data exported from Polygon (xlsx format)."""
    wb = openpyxl.load_workbook(filename)
    ws = wb['Normal']
    colnames = (cell.value for cell in next(ws.rows))  # first row: col names
    normaldata = dict()
    # read the columns and produce dict of numpy arrays
    for colname, col in zip(colnames, ws.columns):
        if colname is None:
            continue
        # pick values from row 4 onwards (skips units etc.)
        data = np.fromiter((c.value for k, c in enumerate(col) if k >= 3), float)
        data = data[~np.isnan(data)]  # drop empty rows
        # rewrite the coordinate
        colname = colname.replace(' (1)', 'X')
        colname = colname.replace(' (2)', 'Y')
        colname = colname.replace(' (3)', 'Z')
        # scalar power variables (ones ending in 'Power')
        # are written as Z component (Nexus convention)
        if colname[-5:] == 'Power':
            colname += 'Z'
        normaldata[colname] = (
            np.stack([normaldata[colname], data], axis=1)
            if colname in normaldata
            else data
        )
    return _check_normaldata(normaldata)


def _write_xlsx(normaldata, filename):
    """Save normal data dict into Polygon xlsx format"""
    repl_di = {'X': ' (1)', 'Y': ' (2)', 'Z': ' (3)'}
    wb = openpyxl.Workbook()
    ws = wb.active
    ws.title = 'Normal'
    for n, var in enumerate(sorted(normaldata), 1):
        logger.debug('writing %s' % var)
        data = normaldata[var]
        nrows, ncols = data.shape
        if nrows not in (51, 1) or ncols != 2:
            raise ValueError(
                'normal data has unexpected dimensions: %d x %d' % (nrows, ncols)
            )
        # convert trailing dimension to number
        if var[-1] in 'XYZ':
            if 'Power' in var:
                var_ = var[:-1]
            else:
                var_ = var[:-1] + repl_di[var[-1]]
        else:
            var_ = var
        firstcol, secondcol = 2 * n - 1, 2 * n
        # write (data min, data max) columns for each variable
        for col, coldata in zip([firstcol, secondcol], [data[:, 0], data[:, 1]]):
            ws.cell(column=col, row=1, value=var_)  # column header
            ws.cell(column=col, row=2, value=0)  # not clear what this is
            ws.cell(
                column=col, row=3, value='unknown'
            )  # supposed to be unit, not used by us
            for k, val in enumerate(coldata):
                ws.cell(column=col, row=4 + k, value=val)
    logger.debug('saving %s' % filename)
    wb.save(filename=filename)


def normals_from_data(data):
    """Compute normaldata from data dict output by stats.collect_trial_data"""
    normaldata = dict()
    for mod in models_all:
        thevars = mod.varlabels_noside
        for var in thevars:
            rvar, lvar = 'R' + var, 'L' + var
            rcurves, lcurves = data[rvar], data[lvar]
            if rcurves is None or lcurves is None:
                logger.warning('cannot get model data for %s' % var)
                continue
            # combine data for L/R
            curves = np.concatenate([rcurves, lcurves])
            # mean and median should coincide for normal distribution but
            # median is less sensitive to outliers, so use it
            curve_med = np.median(curves, axis=0)
            curve_std = numutils.mad(curves, axis=0)
            # traditionally normaldata uses 2% cycle intervals, so downsample
            curve_med_ds = curve_med[::2]
            curve_std_ds = curve_std[::2]
            lower_vardata = curve_med_ds - curve_std_ds
            upper_vardata = curve_med_ds + curve_std_ds
            normaldata[var] = np.stack([lower_vardata, upper_vardata], axis=1)
    return normaldata



# -*- coding: utf-8 -*-
"""

Plotting functionality shared between backends.

@author: Jussi (jnu@iki.fi)
"""


from builtins import next
from past.builtins import basestring
from itertools import cycle
from collections import defaultdict, namedtuple
import datetime
import logging
import numpy as np

from .. import models
from ..config import cfg
from ..envutils import GaitDataError


logger = logging.getLogger(__name__)


def _var_unit(vardef):
    """Return unit for a vardef"""
    varname = vardef[0]
    themodel = models.model_from_var(varname)
    return themodel.units[varname]


def _nested_get(di, keys):
    """Get a value from a nested dict, using a list of keys"""
    for key in keys:
        di = di[key]  # iterate until we exhaust the nested keys
    return di


def _compose_varname(vardef):
    """Compose a variable name for extracted variable.

    E.g. ['HipAnglesX', 'peaks', 'swing', 'max']
    -> 'Hip flexion maximum during swing phase'
    """
    varname = vardef[0]
    # get variable description from gaitutils.models
    themodel = models.model_from_var(varname)
    name = themodel.varlabels_noside[varname]
    if vardef[1] == 'contact':
        name += ' at IC'
    elif vardef[1] in ['peaks', 'extrema']:
        phase = vardef[2]  # swing, stance etc.
        valtype = vardef[3]  # min, max etc.
        val_trans = {'max': 'max.', 'min': 'min.'}
        if phase == 'overall':
            name += ', %s %s' % (phase, val_trans[valtype])
        else:
            name += ', %s phase %s' % (phase, val_trans[valtype])
        if vardef[1] == 'peaks':
            name += ' peak'
    return name


def _get_trial_cycles(trial, cycles, max_cycles):
    """Get the specified cycles from a trial"""
    model_cycles = trial.get_cycles(
        cycles['model'], max_cycles_per_context=max_cycles['model']
    )
    marker_cycles = trial.get_cycles(
        cycles['marker'], max_cycles_per_context=max_cycles['marker']
    )
    emg_cycles = trial.get_cycles(
        cycles['emg'], max_cycles_per_context=max_cycles['emg']
    )
    allcycles = list(set.union(set(model_cycles), set(marker_cycles), set(emg_cycles)))
    if not allcycles:
        logger.debug('trial %s has no cycles of specified type' % trial.trialname)
    logger.debug(
        'got total of %d cycles for %s (%d model, %d marker, %d EMG)'
        % (
            len(allcycles),
            trial.trialname,
            len(model_cycles),
            len(marker_cycles),
            len(emg_cycles),
        )
    )
    Cyclebunch = namedtuple(
        'Cyclebunch', 'allcycles, model_cycles, marker_cycles, emg_cycles'
    )
    return Cyclebunch(allcycles, model_cycles, marker_cycles, emg_cycles)


def _emg_yscale(emg_mode):
    """Compute EMG y range for plotting"""
    if emg_mode == 'rms':
        emg_yrange = np.array([0, cfg.plot.emg_yscale]) * cfg.plot.emg_multiplier
    else:
        emg_yrange = (
            np.array([-cfg.plot.emg_yscale, cfg.plot.emg_yscale])
            * cfg.plot.emg_multiplier
        )
    return emg_yrange


def _color_by_params(spec, mapper, trial, cyc, context, dimension=None):
    """Helper to return a color.

    spec is colorspec and mapper is a color mapper. Returns color according to
    trial, context etc., depending on what spec says. Depending on spec, color
    may come from the mapper or elsewhere (e.g. config). 'dimension' refers to
    variable dimension in case of multidimensional vars (e.g. markers).
    """
    if spec == 'session':
        return mapper[trial.sessiondir]
    elif spec == 'trial':
        return mapper[trial]
    elif spec == 'cycle':
        return mapper[cyc]
    elif spec == 'context':
        return cfg.plot.context_colors[context]
    elif spec == 'dimension':
        return mapper[dimension]
    elif spec is None:
        return mapper[None]
    else:
        raise RuntimeError('Unexpected colorspec: %s' % spec)


def _style_by_params(spec, mapper, trial, cyc, context, dimension=None):
    """Helper to return a style.

    See above for details."""
    if spec == 'session':
        return mapper[trial.sessiondir]
    elif spec == 'trial':
        return mapper[trial]
    elif spec == 'cycle':
        return mapper[cyc]
    elif spec == 'context':
        return mapper[context]
    elif spec == 'dimension':
        return mapper[dimension]
    elif spec is None:
        return mapper[None]
    else:
        raise RuntimeError('Unexpected colorspec: %s' % spec)


def _cyclical_mapper(it):
    """Map iterator to keys cyclically.

    Example:
    colors = ['red', 'blue', 'yellow']
    mapper = _cyclical_mapper(colors)
    mapper['foo']  # red
    mapper['bar']  # blue
    mapper['baz']  # yellow
    mapper['zzz']  # red (iterator cycles over)
    mapper['foo']  # red (old mappings are preserved)
    """
    cyc_it = cycle(it)
    return defaultdict(lambda: next(cyc_it))


def _handle_cyclespec(cycles):
    """Handle cyclespec argument to plotter functions"""
    default_cycles = cfg.plot.default_cycles
    if cycles == 'unnormalized':
        cycles = {vartype: 'unnormalized' for vartype in default_cycles}
    elif cycles is None:
        cycles = default_cycles
    elif isinstance(cycles, dict):
        if set(cycles) - set(default_cycles):  # unknown keys
            raise ValueError('invalid cycle argument')
        _defcycles = default_cycles.copy()
        _defcycles.update(cycles)
        cycles = _defcycles
    else:
        raise ValueError('invalid cycle argument')
    return cycles


def _handle_style_and_color_args(style_by, color_by):
    """Handle style and color choice"""
    vals_ok = set(('session', 'trial', 'context', 'dimension', None))
    style_by_defaults = cfg.plot.style_by
    if style_by is None:
        style_by = dict()
    elif isinstance(style_by, basestring):
        style_by = {'model': style_by}
    elif not isinstance(style_by, dict):
        raise TypeError('style_by must be str or dict')
    for k in set(style_by_defaults) - set(style_by):
        style_by[k] = style_by_defaults[k]  # update missing values
    if not set(style_by.values()).issubset(vals_ok):
        raise ValueError('invalid style_by argument in %s' % style_by.items())

    color_by_defaults = cfg.plot.color_by
    if color_by is None:
        color_by = dict()
    elif isinstance(color_by, basestring):
        color_by = {'model': color_by, 'emg': color_by}
    elif not isinstance(color_by, dict):
        raise TypeError('color_by must be str or dict')
    for k in set(color_by_defaults) - set(color_by):
        color_by[k] = color_by_defaults[k]  # update missing values
    if not set(color_by.values()).issubset(vals_ok):
        raise ValueError('invalid color_by argument in %s' % color_by.items())

    return style_by, color_by


def _style_mpl_to_plotly(style):
    """Map style argument from matplotlib to plotly"""
    return {'-': 'solid', '--': '5px', '-.': 'dashdot', ':': '2px'}[style]


def _var_title(var):
    """Get proper title for a variable"""
    mod = models.model_from_var(var)
    if mod:
        if var in mod.varlabels_noside:
            return mod.varlabels_noside[var]
        elif var in mod.varlabels:
            return mod.varlabels[var]
    elif var in cfg.emg.channel_labels:
        return cfg.emg.channel_labels[var]
    else:
        return var


def _truncate_trialname(trialname):
    """Shorten trial name"""
    tname_split = trialname.split('_')
    try:  # see if trial begins with valid date
        datetxt = '-'.join(tname_split[:3])
        d = datetime.datetime.strptime(datetxt, '%Y-%m-%d')
    except ValueError:
        # if trial was not named as expected, do nothing
        return trialname
    if len(tname_split) >= 5:
        # trialname is probably of the standard form:
        # yyyy_mm_dd_measinfo_sessioninfo1_sessioninfo2_trialn
        measinfo = tname_split[3]
        if len(tname_split) > 5:
            # pick all sessioninfo strings
            sessioninfo = '/'.join(tname_split[4:-1])
        else:
            sessioninfo = ''
        s = '%d/%d %s' % (d.month, d.year, measinfo)
        if sessioninfo:
            s += '/%s' % sessioninfo
        return s
    else:
        return trialname


def _get_cycle_name(trial, cyc, name_type):
    """Return descriptive name for a gait cycle"""
    if name_type == 'name_with_tag':
        cyclename = '%s/%s' % (trial.trialname, trial.eclipse_tag)
    elif name_type == 'short_name_with_tag':
        cyclename = '%s / %s' % (
            _truncate_trialname(trial.trialname),
            trial.eclipse_tag,
        )
    elif name_type == 'short_name_with_tag_and_cycle':
        cyclename = _truncate_trialname(trial.trialname)
        if trial.eclipse_tag is not None:
            cyclename += ' %s/%s' % (trial.eclipse_tag, cyc.name)
    elif name_type == 'tag_only':
        cyclename = trial.eclipse_tag
    elif name_type == 'tag_with_cycle':
        cyclename = '%s/%s' % (trial.eclipse_tag, cyc.name)
    elif name_type == 'full':
        cyclename = '%s/%s' % (trial.name_with_description, cyc.name)
    elif name_type == 'short_name_with_cyclename':
        cyclename = '%s/%s' % (_truncate_trialname(trial.trialname), cyc.name)
    else:
        raise ValueError('Invalid name_type')
    return cyclename


def _triage_var(var, trial):
    """Return category of variable (model, marker etc.).

    Returns 'model', 'marker' or 'emg' for known types,
    'unknown' if type cannot be inferred, None for None
    """
    categs = {'model': False, 'marker': False, 'emg': False}
    if var is None:
        return None
    if models.model_from_var(var):
        categs['model'] = True
    if var in trial._full_marker_data:
        categs['marker'] = True
    if var in cfg.emg.channel_labels:
        categs['emg'] = True
    if list(categs.values()).count(True) > 1:
        categs_matching = [key for key, val in categs.items() if val]
        raise GaitDataError(
            'ambiguous variable name %s (matches categories %s)'
            % (var, categs_matching)
        )
    else:
        for k, v in categs.items():
            if v:
                return k
        return 'unknown'



# -*- coding: utf-8 -*-
"""
layout handling

@author: Jussi (jnu@iki.fi)
"""

from builtins import zip
import logging

from ..config import cfg
from ..envutils import GaitDataError

logger = logging.getLogger(__name__)


def get_layout(layout_name):
    """Get a layout from config.

    Parameters
    ----------
    layout_name : str
        Name of layout.

    Returns
    -------
    list
        The layout.
    """
    # our default layout is PiG lower body kinematics
    if layout_name is None:
        layout_name = 'lb_kinematics'
    try:
        return getattr(cfg.layouts, layout_name)
    except AttributeError:
        raise GaitDataError('No such layout %s' % layout_name)


def _check_layout(layout):
    """Check consistency of layout.

    Returns a tuple (n_of_rows, n_of_cols) if layout is ok.
    Otherwise raises a TypeError.
    """
    if not layout:
        raise TypeError(
            'Empty plotting layout: %s. If EMG layout, check that channels are ok.'
            % layout
        )
    if not isinstance(layout, list):
        raise TypeError('Invalid plotting layout')
    if not isinstance(layout[0], list):
        raise TypeError('Invalid plotting layout')
    nrows = len(layout)
    ncols = len(layout[0])
    if ncols < 1:
        raise TypeError('Invalid plotting layout: %s' % layout)
    for col in layout:
        if not isinstance(col, list) or len(col) != ncols:
            raise TypeError('Inconsistent plotting layout: %s' % layout)
    return nrows, ncols


def _rm_dead_channels(emgs, layout):
    """From EMG layout, remove rows that do not have any good data"""
    if not isinstance(emgs, list):
        emgs = [emgs]
    chs_ok = None
    for i, emg in enumerate(emgs):
        # accept channels w/ status ok, or anything that is NOT a
        # preconfigured EMG channel
        chs_ok_ = [
            ch not in cfg.emg.channel_labels or emg.status_ok(ch)
            for row in layout
            for ch in row
        ]
        # previously OK chs propagated as ok
        chs_ok = [a or b for a, b in zip(chs_ok, chs_ok_)] if i > 0 else chs_ok_
    if not chs_ok:
        raise GaitDataError('No acceptable channels in any of the EMGs')
    rowlen = len(layout[0])
    lout = list(zip(*[iter(chs_ok)] * rowlen))  # grouper recipe from itertools
    rows_ok = [any(row) for row in lout]
    layout = [row for i, row in enumerate(layout) if rows_ok[i]]
    return layout



import logging
import os
import sys

from . import (
    c3d,
    config,
    eclipse,
    emg,
    envutils,
    gui,
    models,
    nexus,
    normaldata,
    numutils,
    read_data,
    report,
    sessionutils,
    stats,
    timedist,
    trial,
    utils,
    videos,
    viz,
)

# in case we want to print stuff from config.py, it's better to delay
# the import until this point (after the stdout fix above)
from .config import cfg
from .envutils import GaitDataError

# the main purpose of adding the null handler is to disable the
# default stderr logging for levels >= warning, which may be
# annoying in some cases
logging.getLogger('gaitutils').addHandler(logging.NullHandler())


def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


# fake stdout and stderr not being available if run
# under pythonw.exe on Windows
if (
    sys.platform.find('win') != -1
    and sys.executable.find('pythonw') != -1
    and not run_from_ipython()
):
    blackhole = open(os.devnull, 'w')
    sys.stdout = sys.stderr = blackhole



# -*- coding: utf-8 -*-
"""
Create web-based gait report using dash.

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division

from builtins import str
from past.builtins import basestring
import plotly.graph_objs as go
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
import flask
from flask import request
from collections import OrderedDict
import logging
import os.path as op
import base64

from ulstools.num import age_from_hetu

from .. import (
    normaldata,
    models,
    sessionutils,
    numutils,
    videos,
)
from ..config import cfg
from ..envutils import GaitDataError
from ..sessionutils import enf_to_trialfile
from ..trial import Trial
from ..viz.plot_plotly import plot_trials, plot_extracted_box
from ..viz import timedist, layouts
from ..stats import AvgTrial, _trials_extract_values


# Py2: for Python 3 compatibility
try:
    import cPickle as pickle
except ImportError:
    import pickle


logger = logging.getLogger(__name__)


def _make_dropdown_lists(options):
    """Helper for dcc.Dropdown.

    Take a list of label/value dicts (with arbitrary type values) and returns
    (list, dict). Needed since dcc.Dropdown can only take str values. identity
    is fed to dcc.Dropdown() and mapper is used for getting the actual values at
    the callback."""
    identity = list()
    mapper = dict()
    for option in options:
        di = {'label': option['label'], 'value': option['label']}
        if 'disabled' in option and option['disabled']:
            di['disabled'] = True
        identity.append(di)
        mapper[option['label']] = option['value']
    return identity, mapper


def _shutdown_server():
    """Shutdown flask server, see http://flask.pocoo.org/snippets/67/"""
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


def _report_name(sessions, long_name=True):
    """Create a title for the dash report"""
    sessions_str = ' / '.join(op.split(s)[-1] for s in sessions)
    if long_name:
        report_type = (
            'Single session report' if len(sessions) == 1 else 'Comparison report'
        )
    else:
        report_type = 'Single' if len(sessions) == 1 else 'Comparison'
    return '%s: %s' % (report_type, sessions_str)


def dash_report(
    sessions,
    info=None,
    tags=None,
    signals=None,
    recreate_plots=None,
    video_only=None,
):
    """Create a gait report dash app.

    Parameters
    ----------
    sessions : list
        List of session directories. Up to three sessions can be compared in the
        report.
    info : dict | None
        The patient info.
    tags : list | None
        Eclipse tags for finding dynamic gait trials. If None, will be taken from config.
    signals : ProgressSignals | None
        Instance of ProgressSignals, used to send progress updates across
        threads and track cancel flag. If None, a dummy one will be created.
    recreate_plots : bool
        If True, force recreation of report figures. Otherwise, figures cached
        to disk will be used, unless the report c3d files have changed (a
        checksum mechanism is used to verify this).
    video_only : bool
        If True, create a video-only report.

    Returns
    -------
    dash.Dash | None
        The dash (flask) app, or None if report creation failed.
    """

    if recreate_plots is None:
        recreate_plots = False

    if video_only is None:
        video_only = False

    # relative width of left panel (1-12)
    # uncomment to use narrower video panel for 3-session comparison
    # LEFT_WIDTH = 8 if len(sessions) == 3 else 7
    LEFT_WIDTH = 8
    VIDS_TOTAL_HEIGHT = 88  # % of browser window height

    if len(sessions) < 1 or len(sessions) > 3:
        raise ValueError('Need a list of one to three sessions')
    is_comparison = len(sessions) > 1
    report_name = _report_name(sessions)
    info = info or sessionutils.default_info()

    # tags for dynamic trials
    if tags is None:
        dyn_tags = cfg.eclipse.tags
    else:
        dyn_tags = tags

    # this tag will be shown in the menu for static trials
    static_tag = 'Static'

    # get the camera labels
    # reduce to a set, since there may be several labels for given id
    camera_labels = set(cfg.general.camera_labels.values())
    # add camera labels for overlay videos
    # XXX: may cause trouble if camera labels already contain the string 'overlay'
    camera_labels_overlay = [lbl + ' overlay' for lbl in camera_labels]
    camera_labels.update(camera_labels_overlay)
    # build dict of videos for given tag / camera label
    # videos will be listed in session order
    vid_urls = dict()
    all_tags = dyn_tags + [static_tag] + cfg.eclipse.video_tags
    for tag in all_tags:
        vid_urls[tag] = dict()
        for camera_label in camera_labels:
            vid_urls[tag][camera_label] = list()

    # collect all session enfs into dict
    enfs = {session: dict() for session in sessions}
    data_enfs = list()  # enfs that are used for data
    signals.progress.emit('Collecting trials...', 0)
    for session in sessions:
        if signals.canceled:
            return None
        enfs[session] = dict(dynamic=dict(), static=dict(), vid_only=dict())
        # collect dynamic trials for each tag
        for tag in dyn_tags:
            dyns = sessionutils.get_enfs(session, tags=tag, trial_type='dynamic')
            if len(dyns) > 1:
                logger.warning('multiple tagged trials (%s) for %s' % (tag, session))
            dyn_trial = dyns[-1:]
            enfs[session]['dynamic'][tag] = dyn_trial  # may be empty list
            if dyn_trial:
                data_enfs.extend(dyn_trial)
        # require at least one dynamic trial for each session
        if not any(enfs[session]['dynamic'][tag] for tag in dyn_tags):
            raise GaitDataError('No tagged dynamic trials found for %s' % (session))
        # collect static trial (at most 1 per session)
        # -prefer enfs that have a corresponding c3d file, even for a video-only report
        # (so that the same static gets used for both video-only and full reports)
        # -prefer the newest enf file
        sts = sessionutils.get_enfs(session, trial_type='static')
        for st in reversed(sts):  # newest first
            st_c3d = sessionutils.enf_to_trialfile(st, '.c3d')
            if op.isfile(st_c3d):
                static_trial = [st]
                break
        else:
            # no c3ds were found - just pick the latest static trial
            static_trial = sts[-1:]
        enfs[session]['static'][static_tag] = static_trial
        if static_trial:
            data_enfs.extend(static_trial)
        # collect video-only dynamic trials
        for tag in cfg.eclipse.video_tags:
            dyn_vids = sessionutils.get_enfs(session, tags=tag)
            if len(dyn_vids) > 1:
                logger.warning(
                    'multiple tagged video-only trials (%s) for %s' % (tag, session)
                )
            enfs[session]['vid_only'][tag] = dyn_vids[-1:]

    # collect all videos for given tag and camera, listed in session order
    signals.progress.emit('Finding videos...', 0)
    for session in sessions:
        for trial_type in enfs[session]:
            for tag, enfs_this in enfs[session][trial_type].items():
                if enfs_this:
                    enf = enfs_this[0]  # only one enf per tag and session
                    for camera_label in camera_labels:
                        overlay = 'overlay' in camera_label
                        real_camera_label = (
                            camera_label[: camera_label.find(' overlay')]
                            if overlay
                            else camera_label
                        )
                        # need to convert filename, since get_trial_videos cannot
                        # deal with enf names
                        c3d = enf_to_trialfile(enf, 'c3d')
                        vids_this = videos.get_trial_videos(
                            c3d,
                            camera_label=real_camera_label,
                            vid_ext='.ogv',
                            overlay=overlay,
                        )
                        if vids_this:
                            vid = vids_this[0]
                            url = '/static/%s' % op.split(vid)[1]
                            vid_urls[tag][camera_label].append(url)

    # build dcc.Dropdown options list for cameras and tags
    # list cameras which have videos for any tag
    opts_cameras = list()
    for camera_label in sorted(camera_labels):
        if any(vid_urls[tag][camera_label] for tag in all_tags):
            opts_cameras.append({'label': camera_label, 'value': camera_label})
    # list tags which have videos for any camera
    opts_tags = list()
    for tag in all_tags:
        if any(vid_urls[tag][camera_label] for camera_label in camera_labels):
            opts_tags.append({'label': '%s' % tag, 'value': tag})
    # add null entry in case we got no videos at all
    if not opts_tags:
        opts_tags.append({'label': 'No videos', 'value': 'no videos', 'disabled': True})

    # create (or load) the figures
    # this section is only run if we have c3d data
    if not video_only:
        data_c3ds = [enf_to_trialfile(enffile, 'c3d') for enffile in data_enfs]
        # at this point, all the c3ds need to exist
        missing = [fn for fn in data_c3ds if not op.isfile(fn)]
        if missing:
            missing_trials = ', '.join(
                [op.splitext(op.split(fn)[-1])[0] for fn in missing]
            )
            raise GaitDataError(
                'C3D files are missing for following trials: %s' % missing_trials
            )
        # see whether we can load report figures from disk
        digest = numutils._files_digest(data_c3ds)
        logger.debug('report data digest: %s' % digest)
        # data is always saved into alphabetically first session
        data_dir = sorted(sessions)[0]
        data_fn = op.join(data_dir, 'web_report_%s.dat' % digest)
        if op.isfile(data_fn) and not recreate_plots:
            logger.debug('loading saved report data from %s' % data_fn)
            signals.progress.emit('Loading saved report...', 0)
            with open(data_fn, 'rb') as f:
                saved_report_data = pickle.load(f)
        else:
            saved_report_data = dict()
            logger.debug('no saved data found or recreate forced')

        # make Trial instances for all dynamic and static trials
        # this is currently necessary even if saved figures are used
        trials_dyn = list()
        trials_dyn_dict = OrderedDict()  # also organize dynamic trials by session
        trials_static = list()
        for session in sessions:
            trials_dyn_dict[session] = list()
            for tag in dyn_tags:
                if enfs[session]['dynamic'][tag]:
                    if signals.canceled:
                        return None
                    c3dfile = enf_to_trialfile(enfs[session]['dynamic'][tag][0], 'c3d')
                    tri = Trial(c3dfile)
                    trials_dyn.append(tri)
                    trials_dyn_dict[session].append(tri)
            if enfs[session]['static'][static_tag]:
                c3dfile = enf_to_trialfile(enfs[session]['static']['Static'][0], 'c3d')
                tri = Trial(c3dfile)
                trials_static.append(tri)

        emg_auto_layout = None

        # stuff that's needed to (re)create the figures
        if not saved_report_data:
            age = None
            if info['hetu'] is not None:
                # compute subject age at session time
                session_dates = [
                    sessionutils.get_session_date(session) for session in sessions
                ]
                ages = [age_from_hetu(info['hetu'], d) for d in session_dates]
                age = max(ages)

            # create Markdown text for patient info
            patient_info_text = '##### %s ' % (
                info['fullname'] if info['fullname'] else 'Name unknown'
            )
            if info['hetu']:
                patient_info_text += '(%s)' % info['hetu']
            patient_info_text += '\n\n'
            # if age:
            #     patient_info_text += 'Age at measurement time: %d\n\n' % age

            # load normal data for gait models; we have to do it here instead of
            # leaving it up to plot_trials, since it's session (age) specific
            signals.progress.emit('Loading normal data...', 0)
            model_normaldata = normaldata._read_configured_model_normaldata(age)

            # make average trials for each session
            avg_trials = [
                AvgTrial.from_trials(trials_dyn_dict[session], sessionpath=session)
                for session in sessions
            ]

            # prepare for the curve-extracted value plots
            logger.debug('extracting values for curve-extracted plots...')
            # Py2: the tuple->OrderedDict trick is used in pdf.py to preserve the
            # ordering of plots under Py2; it isn't really be necessary here
            # since page_layouts defines the order
            vardefs_dict = OrderedDict(cfg.report.vardefs)
            allvars = [
                vardef[0] for vardefs in vardefs_dict.values() for vardef in vardefs
            ]
            from_models = set(models.model_from_var(var) for var in allvars)
            if None in from_models:
                raise GaitDataError('unknown variables in extract list: %s' % allvars)
            curve_vals = OrderedDict(
                [
                    (
                        op.split(session)[-1],
                        _trials_extract_values(trials, from_models=from_models),
                    )
                    for session, trials in trials_dyn_dict.items()
                ]
            )

            # in EMG layout, keep chs that are active in any of the trials
            signals.progress.emit('Reading EMG data', 0)
            try:
                emgs = [tr.emg for tr in trials_dyn]
                emg_auto_layout = layouts._rm_dead_channels(emgs, cfg.layouts.std_emg)
                if not emg_auto_layout:
                    emg_auto_layout = None
            except GaitDataError:
                emg_auto_layout = None

        # the layouts are specified as lists of tuples: (title, layout_spec)
        # where title is the page title, and layout_spec is either string or tuple.
        # if string, it denotes a special layout (e.g. 'patient_info')
        # if tuple, the first element should be the string 'layout_name' and the second
        # a gaitutils configured layout name;
        # alternatively the first element can be 'layout' and the second element a
        # valid gaitutils layout
        # Py2: we need to use tuple->OrderedDict to preserve the ordering
        page_layouts = OrderedDict(cfg.web_report.page_layouts)

        # pick desired single variables from model and append
        # Py2: dict merge below can be done more elegantly once Py2 is dropped
        pig_singlevars_ = models.pig_lowerbody.varlabels_noside.copy()
        pig_singlevars_.update(models.pig_lowerbody_kinetics.varlabels_noside)
        pig_singlevars = sorted(pig_singlevars_.items(), key=lambda item: item[1])
        singlevars = OrderedDict(
            [(varlabel, ('layout', [[var]])) for var, varlabel in pig_singlevars]
        )
        page_layouts.update(singlevars)

        # add supplementary data for normal layouts
        supplementary_default = dict()

        dd_opts_multi_upper = list()
        dd_opts_multi_lower = list()

        # loop through the layouts, create or load figures
        report_data_new = dict()
        for k, (page_label, layout_spec) in enumerate(page_layouts.items()):
            signals.progress.emit(
                'Creating plot: %s' % page_label, 100 * k / len(page_layouts)
            )
            if signals.canceled:
                return None
            # for comparison report, include session info in plot legends and
            # use session specific line style
            emg_mode = None
            if is_comparison:
                legend_type = cfg.report.comparison_legend_type
                style_by = cfg.report.comparison_style_by
                color_by = cfg.report.comparison_color_by
                if cfg.report.comparison_emg_rms:
                    emg_mode = 'rms'
            else:
                legend_type = cfg.report.legend_type
                style_by = cfg.report.style_by
                color_by = cfg.report.color_by

            try:
                if saved_report_data:
                    logger.debug('loading %s from saved report data' % page_label)
                    if page_label not in saved_report_data:
                        # will be caught, resulting in empty menu item
                        raise RuntimeError
                    else:
                        figdata = saved_report_data[page_label]
                else:
                    logger.debug('creating figure data for %s' % page_label)
                    # the 'special' layouts are indicated by a string
                    if isinstance(layout_spec, basestring):
                        if layout_spec == 'time_dist':
                            figdata = timedist.plot_comparison(
                                sessions, big_fonts=False, backend='plotly'
                            )
                        elif layout_spec == 'patient_info':
                            figdata = patient_info_text
                        elif layout_spec == 'static_kinematics':
                            layout_ = cfg.layouts.lb_kinematics
                            figdata = plot_trials(
                                trials_static,
                                layout_,
                                model_normaldata=False,
                                cycles='unnormalized',
                                legend_type='short_name_with_cyclename',
                                style_by=style_by,
                                color_by=color_by,
                                big_fonts=True,
                            )
                        elif layout_spec == 'static_emg':
                            layout_ = cfg.layouts.std_emg
                            figdata = plot_trials(
                                trials_static,
                                layout_,
                                model_normaldata=False,
                                cycles='unnormalized',
                                legend_type='short_name_with_cyclename',
                                style_by=style_by,
                                color_by=color_by,
                                big_fonts=True,
                            )
                        elif layout_spec == 'emg_auto' and emg_auto_layout is not None:
                            figdata = plot_trials(
                                trials_dyn,
                                emg_auto_layout,
                                emg_mode=emg_mode,
                                legend_type=legend_type,
                                style_by=style_by,
                                color_by=color_by,
                                supplementary_data=supplementary_default,
                                big_fonts=True,
                            )
                        elif layout_spec == 'kinematics_average':
                            layout_ = cfg.layouts.lb_kinematics
                            figdata = plot_trials(
                                avg_trials,
                                layout_,
                                style_by=style_by,
                                color_by=color_by,
                                model_normaldata=model_normaldata,
                                big_fonts=True,
                            )
                        elif layout_spec == 'disabled':
                            # will be caught, resulting in empty menu item
                            raise RuntimeError
                        else:  # unrecognized layout; this is not caught by us
                            raise Exception(
                                'Invalid page layout: %s' % str(layout_spec)
                            )

                    # regular layouts and curve-extracted layouts are indicated by tuple
                    elif isinstance(layout_spec, tuple):
                        if layout_spec[0] in ['layout_name', 'layout']:
                            if layout_spec[0] == 'layout_name':
                                # get a configured layout by name
                                layout = layouts.get_layout(layout_spec[1])
                            else:
                                # it's already a valid layout
                                layout = layout_spec[1]
                            # plot according to layout
                            figdata = plot_trials(
                                trials_dyn,
                                layout,
                                model_normaldata=model_normaldata,
                                emg_mode=emg_mode,
                                legend_type=legend_type,
                                style_by=style_by,
                                color_by=color_by,
                                supplementary_data=supplementary_default,
                                big_fonts=True,
                            )
                        elif layout_spec[0] == 'curve_extracted':
                            the_vardefs = vardefs_dict[layout_spec[1]]
                            figdata = plot_extracted_box(curve_vals, the_vardefs)
                        else:
                            raise Exception(
                                'Invalid page layout: %s' % str(layout_spec)
                            )
                    else:
                        raise Exception('Invalid page layout: %s' % str(layout_spec))

                # save the newly created data
                if not saved_report_data:
                    if isinstance(figdata, go.Figure):
                        # serialize go.Figures before saving
                        # this makes them much faster for pickle to handle
                        # apparently dcc.Graph can eat the serialized json directly,
                        # so no need to do anything on load
                        figdata_ = figdata.to_plotly_json()
                    else:
                        figdata_ = figdata
                    report_data_new[page_label] = figdata_

                # make the upper and lower panel graphs from figdata, depending
                # on data type
                def _is_base64(s):
                    try:
                        return base64.b64encode(base64.b64decode(s)) == s
                    except Exception:
                        return False

                # this is for old style timedist figures that were in base64
                # encoded svg
                if layout_spec == 'time_dist' and _is_base64(figdata):
                    graph_upper = html.Img(
                        src='data:image/svg+xml;base64,{}'.format(figdata),
                        id='gaitgraph%d' % k,
                        style={'height': '100%'},
                    )
                    graph_lower = html.Img(
                        src='data:image/svg+xml;base64,{}'.format(figdata),
                        id='gaitgraph%d' % (len(page_layouts) + k),
                        style={'height': '100%'},
                    )
                elif layout_spec == 'patient_info':
                    graph_upper = dcc.Markdown(figdata)
                    graph_lower = graph_upper
                else:
                    # plotly fig -> dcc.Graph
                    graph_upper = dcc.Graph(
                        figure=figdata, id='gaitgraph%d' % k, style={'height': '100%'}
                    )
                    graph_lower = dcc.Graph(
                        figure=figdata,
                        id='gaitgraph%d' % (len(page_layouts) + k),
                        style={'height': '100%'},
                    )
                dd_opts_multi_upper.append({'label': page_label, 'value': graph_upper})
                dd_opts_multi_lower.append({'label': page_label, 'value': graph_lower})

            except (RuntimeError, GaitDataError) as e:  # could not create a figure
                logger.warning(u'failed to create figure for %s: %s' % (page_label, e))
                # insert the menu options but make them disabled
                dd_opts_multi_upper.append(
                    {'label': page_label, 'value': page_label, 'disabled': True}
                )
                dd_opts_multi_lower.append(
                    {'label': page_label, 'value': page_label, 'disabled': True}
                )
                continue

        opts_multi, mapper_multi_upper = _make_dropdown_lists(dd_opts_multi_upper)
        opts_multi, mapper_multi_lower = _make_dropdown_lists(dd_opts_multi_lower)

        # if plots were newly created, save them to disk
        if not saved_report_data:
            logger.debug('saving report data into %s' % data_fn)
            signals.progress.emit('Saving report data to disk...', 99)
            with open(data_fn, 'wb') as f:
                pickle.dump(report_data_new, f, protocol=-1)

    def make_left_panel(split=True, upper_value='Kinematics', lower_value='Kinematics'):
        """Helper to make the left graph panels. If split=True, make two stacked panels"""

        # the upper graph & dropdown
        items = [
            dcc.Dropdown(
                id='dd-vars-upper-multi',
                clearable=False,
                options=opts_multi,
                value=upper_value,
            ),
            html.Div(
                id='div-upper', style={'height': '50%'} if split else {'height': '100%'}
            ),
        ]

        if split:
            # add the lower one
            items.extend(
                [
                    dcc.Dropdown(
                        id='dd-vars-lower-multi',
                        clearable=False,
                        options=opts_multi,
                        value=lower_value,
                    ),
                    html.Div(id='div-lower', style={'height': '50%'}),
                ]
            )

        return html.Div(items, style={'height': '80vh'})

    # create the app
    app = dash.Dash('gaitutils')
    # use local packaged versions of JavaScript libs etc. (no internet needed)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True
    app.title = _report_name(sessions, long_name=False)

    # this is for generating the classnames in the CSS
    num2words = {
        1: 'one',
        2: 'two',
        3: 'three',
        4: 'four',
        5: 'five',
        6: 'six',
        7: 'seven',
        8: 'eight',
        9: 'nine',
        10: 'ten',
        11: 'eleven',
        12: 'twelve',
    }
    classname_left = '%s columns' % num2words[LEFT_WIDTH]
    classname_right = '%s columns' % num2words[12 - LEFT_WIDTH]

    if video_only:
        app.layout = html.Div(
            [  # row
                html.Div(
                    [  # single main div
                        dcc.Dropdown(
                            id='dd-camera',
                            clearable=False,
                            options=opts_cameras,
                            value='Front camera',
                        ),
                        dcc.Dropdown(
                            id='dd-video-tag',
                            clearable=False,
                            options=opts_tags,
                            value=opts_tags[0]['value'],
                        ),
                        html.Div(id='videos'),
                    ],
                    className='12 columns',
                ),
            ],
            className='row',
        )
    else:  # the two-panel layout with graphs and video
        app.layout = html.Div(
            [  # row
                html.Div(
                    [  # left main div
                        html.H6(report_name),
                        dcc.Checklist(
                            id='split-left',
                            options=[{'label': 'Two panels', 'value': 'split'}],
                            value=[],
                        ),
                        # need split=True so that both panels are in initial layout
                        html.Div(make_left_panel(split=True), id='div-left-main'),
                    ],
                    className=classname_left,
                ),
                html.Div(
                    [  # right main div
                        dcc.Dropdown(
                            id='dd-camera',
                            clearable=False,
                            options=opts_cameras,
                            value='Front camera',
                        ),
                        dcc.Dropdown(
                            id='dd-video-tag',
                            clearable=False,
                            options=opts_tags,
                            value=opts_tags[0]['value'],
                        ),
                        html.Div(id='videos'),
                    ],
                    className=classname_right,
                ),
            ],
            className='row',
        )

        @app.callback(
            Output('div-left-main', 'children'),
            [Input('split-left', 'value')],
            [State('dd-vars-upper-multi', 'value')],
        )
        def update_panel_layout(split_panels, upper_value):
            split = 'split' in split_panels
            return make_left_panel(split, upper_value=upper_value)

        @app.callback(
            Output('div-upper', 'children'), [Input('dd-vars-upper-multi', 'value')]
        )
        def update_contents_upper_multi(sel_var):
            return mapper_multi_upper[sel_var]

        @app.callback(
            Output('div-lower', 'children'), [Input('dd-vars-lower-multi', 'value')]
        )
        def update_contents_lower_multi(sel_var):
            return mapper_multi_lower[sel_var]

    def _video_elem(title, url, max_height):
        """Create a video element with title"""
        if not url:
            return 'No video found'
        vid_el = html.Video(
            src=url,
            controls=True,
            loop=True,
            preload='auto',
            title=title,
            style={'max-height': max_height, 'max-width': '100%'},
        )
        # return html.Div([title, vid_el])  # titles above videos
        return vid_el

    @app.callback(
        Output('videos', 'children'),
        [Input('dd-camera', 'value'), Input('dd-video-tag', 'value')],
    )
    def update_videos(camera_label, tag):
        """Create a list of video divs according to camera and tag selection"""
        if tag == 'no videos':
            return 'No videos found'
        vid_urls_ = vid_urls[tag][camera_label]
        if not vid_urls_:
            return 'No videos found'
        nvids = len(vid_urls_)
        max_height = str(int(VIDS_TOTAL_HEIGHT / nvids)) + 'vh'
        return [_video_elem('video', url, max_height) for url in vid_urls_]

    # add a static route to serve session data. be careful outside firewalls
    @app.server.route('/static/<resource>')
    def serve_file(resource):
        for session in sessions:
            filepath = op.join(session, resource)
            if op.isfile(filepath):
                return flask.send_from_directory(session, resource)
        return None

    # add shutdown method - see http://flask.pocoo.org/snippets/67/
    @app.server.route('/shutdown')
    def shutdown():
        logger.debug('Received shutdown request...')
        _shutdown_server()
        return 'Server shutting down...'

    # inject some info of our own
    app._gaitutils_report_name = report_name

    # XXX: the Flask app ends up with a logger by the name of 'gaitutils', which has a default
    # stderr handler. since logger hierarchy corresponds to package hierarchy,
    # this creates a bug where all gaitutils package loggers propagate their messages into
    # the app logger and they get shown multiple times. as a dirty fix, we disable the
    # handlers for the app logger (they still get shown since they propagate to the root logger)
    app.logger.handlers = []

    return app



# -*- coding: utf-8 -*-
"""
Vicon Nexus utils & data readers.

NB: do not use the data readers from this file directly. They are intended to be
called via the read_data module.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function, division
from collections import defaultdict
import sys
import numpy as np
import os.path as op
import psutil
import glob
import logging
import time
import multiprocessing

from .numutils import _change_coords, _isfloat, _rigid_body_extrapolate_markers
from .utils import TrialEvents
from .envutils import GaitDataError
from .config import cfg


logger = logging.getLogger(__name__)


def _find_nexus_path(vicon_path=None):
    """Return path to most recent Nexus version.

    vicon_path is the Vicon root directory.
    """
    if vicon_path is None:
        vicon_path = r'C:\Program Files (x86)\Vicon'  # educated guess
    if not op.isdir(vicon_path):
        return None
    nexus_glob = op.join(vicon_path, 'Nexus?.*')
    nexus_dirs = glob.glob(nexus_glob)
    if not nexus_dirs:
        return None
    nexus_vers = [op.split(dir_)[1][5:] for dir_ in nexus_dirs]
    # convert into major,minor lists: [[2,1], [2,10]] etc.
    try:
        nexus_vers = [[int(s) for s in v.split('.')] for v in nexus_vers]
    except ValueError:
        return None
    # 2-key sort using first major and then minor version number
    idx = nexus_vers.index(max(nexus_vers, key=lambda l: (l[0], l[1])))
    return nexus_dirs[idx]


def _add_nexus_path(vicon_path):
    """Add Nexus SDK dir to sys.path"""

    nexus_path = _find_nexus_path(vicon_path)
    if nexus_path is None:
        logger.warning(
            'cannot locate Nexus installation directory under %s' % vicon_path
        )
        return

    sdk_path = op.join(nexus_path, 'SDK', 'Python')
    if sdk_path not in sys.path:
        sys.path.append(sdk_path)
    else:
        logger.debug('%s already in sys.path' % sdk_path)

    # import from Win32 or Win64 according to bitness of Python interpreter
    bitness = '64' if sys.maxsize > 2 ** 32 else '32'
    win = 'Win' + bitness
    _win_sdk_path = op.join(nexus_path, 'SDK', win)

    # check that the path for the wrong architecture has not already been
    # added to path (this may happen when running inside Nexus)
    win_other = 'Win32' if win == 'Win64' else 'Win64'
    _win_sdk_other = op.join(nexus_path, 'SDK', win_other)
    if _win_sdk_other in sys.path:
        logger.debug('%s already in sys.path, removing' % _win_sdk_other)
        sys.path.remove(_win_sdk_other)

    if _win_sdk_path not in sys.path:
        logger.debug('using Nexus SDK from %s' % _win_sdk_path)
        sys.path.append(_win_sdk_path)
    else:
        logger.debug('%s already in sys.path' % _win_sdk_path)
    return _win_sdk_path


# try to add Nexus SDK to sys.path and import ViconNexus
if sys.version_info.major >= 3:
    logger.debug('running on Python 3 or newer, cannot import Nexus API (yet)')
    nexus_path = ''
else:
    vicon_path = op.normpath(cfg.general.vicon_path)
    nexus_path = _add_nexus_path(vicon_path)
    try:
        import ViconNexus
    except ImportError:
        logger.debug('cannot import Vicon Nexus SDK')
        nexus_path = ''
sys.stdout.flush()  # make sure import warnings get printed


def _nexus_pid():
    """Try to return the PID of the currently running Nexus process"""
    PROCNAME = "Nexus.exe"
    for proc in psutil.process_iter():
        try:
            if proc.name() == PROCNAME:
                return proc.pid
        # catch NoSuchProcess for procs that disappear inside loop
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass
    return None


def _nexus_version():
    """Try to return the actual version of the running Nexus process
    (API does not do that). Hackish and probably unreliable. Returns dict of
    major and minor version number if successful, otherwise (None, None)"""
    PROCNAME = "Nexus.exe"
    for proc in psutil.process_iter():
        try:
            if proc.name() == PROCNAME:
                exname = proc.exe()
                vstart = exname.find('2.')  # assumes ver >2.
                vend = exname.find('\\Nexus.exe')
                if vstart == -1 or vend == -1:
                    return None, None
                try:
                    ver_str = exname[vstart:vend]
                    vmaj, vmin = ver_str.split('.')
                    return int(vmaj), int(vmin)
                except ValueError:  # cannot interpret version string
                    return None, None
        except psutil.AccessDenied:
            pass
    return None, None


def _nexus_ver_greater(major, minor):
    """Checks if running Nexus version is at least the given version"""
    vmaj, vmin = _nexus_version()
    if vmaj is None:
        return False
    else:
        return vmaj >= major and vmin >= minor


def viconnexus():
    """Return a ViconNexus() (SDK control object) instance.

    Raises an exception if Nexus is not running.

    Returns
    -------
    ViconNexus
        The instance.
    """
    _check_nexus()
    return ViconNexus.ViconNexus()


def _close_trial():
    """Try to close currently opened Nexus trial"""
    vicon = viconnexus()
    # this op was not supported before Nexus 2.8
    if _nexus_ver_greater(2, 8):
        logger.info('force closing open trial')
        vicon.CloseTrial(5000)
    else:
        logger.info('current Nexus API version does not support closing trials')


def _open_trial(trialpath, close_first=True):
    """Open trial in Nexus"""
    vicon = viconnexus()
    if close_first:
        _close_trial()
    # Nexus wants the path without filename extension (e.g. .c3d)
    trialpath_ = op.splitext(trialpath)[0]
    vicon.OpenTrial(trialpath_, 60)


def get_subjectnames(single_only=True):
    """Get current subject name(s) from Nexus.

    Parameters
    ----------
    single_only : bool, optional
        Accept and return a single subject only. If True, an exception will be
        raised if Nexus has multiple subjects defined.

    Returns
    -------
    str | list
        The subject name, or a list of names.
    """
    vicon = viconnexus()
    get_sessionpath()  # check whether we can get data
    names_ = vicon.GetSubjectNames()
    if not names_:
        raise GaitDataError('No subject defined in Nexus')
    if single_only:
        if len(names_) > 1:
            raise GaitDataError('Nexus returns multiple subjects')
    # workaround a Nexus 2.6 bug (?) that creates extra names with weird unicode
    # strings
    names_ = [name for name in names_ if u'\ufffd1' not in name]
    return names_[0] if single_only else names_


def _check_nexus():
    """Check whether Nexus is currently running"""
    if not _nexus_pid():
        raise GaitDataError('Vicon Nexus does not seem to be running')


def get_sessionpath():
    """Get path to current Nexus session.

    Returns
    -------
    str
        The path.
    """
    try:
        vicon = viconnexus()
        sessionpath = vicon.GetTrialName()[0]
    except IOError:  # may be raised if Nexus was just terminated
        sessionpath = None
    if not sessionpath:
        raise GaitDataError(
            'Cannot get Nexus session path, no session or maybe in Live mode?'
        )
    return op.normpath(sessionpath)


def _run_pipeline(pipeline, foo, timeout):
    """Wrapper needed for multiprocessing module due to pickle limitations"""
    vicon = viconnexus()
    return vicon.Client.RunPipeline(pipeline, foo, timeout)


def _run_pipelines(pipelines):
    """Run given Nexus pipeline(s).

    Note: this version will stall the calling Python interpreter until the
    pipeline is finished.
    """
    if type(pipelines) != list:
        pipelines = [pipelines]
    for pipeline in pipelines:
        logger.debug('running pipeline: %s' % pipeline)
        result = _run_pipeline(pipeline.encode('utf-8'), '', cfg.autoproc.nexus_timeout)
        if result.Error():
            logger.warning('error while trying to run Nexus pipeline: %s' % pipeline)


def _run_pipelines_multiprocessing(pipelines):
    """Run given Nexus pipeline(s) via the multiprocessing module.

    The idea is to work around the Python global interpreter lock, since the
    Nexus SDK does not release it. By starting a new interpreter process for the
    pipeline, this version causes the invoking thread to sleep and release the
    GIL while the pipeline is running.
    """
    if type(pipelines) != list:
        pipelines = [pipelines]
    for pipeline in pipelines:
        logger.debug('running pipeline via multiprocessing module: %s' % pipeline)
        args = (pipeline.encode('utf-8'), '', cfg.autoproc.nexus_timeout)
        p = multiprocessing.Process(target=_run_pipeline, args=args)
        p.start()
        while p.exitcode is None:
            time.sleep(0.1)


def _get_trialname():
    """Get current Nexus trialname without the session path"""
    vicon = viconnexus()
    trialname_ = vicon.GetTrialName()
    return trialname_[1]


def _is_vicon_instance(obj):
    """Check if obj is an instance of ViconNexus"""
    return obj.__class__.__name__ == 'ViconNexus'


def _get_nexus_subject_param(vicon, name, param):
    """Wrapper to get subject parameter from Nexus."""
    value = vicon.GetSubjectParam(name, param)
    # for unknown reasons, above method may return tuple or float
    # depending on whether script is run from Nexus or outside
    if type(value) == tuple:
        value = value[0] if value[1] else None
    return value


def _get_marker_names(vicon, trajs_only=True):
    """Return marker names from Nexus.

    If trajs_only, only return markers with trajectories.
    """
    subjname = get_subjectnames()
    markers = vicon.GetMarkerNames(subjname)
    # only get markers with trajectories - excludes calibration markers
    if trajs_only:
        markers = [mkr for mkr in markers if vicon.HasTrajectory(subjname, mkr)]
    return markers


def _get_metadata(vicon):
    """Read trial and subject metadata from Nexus.

    See read.data.get_metadata for details."""
    _check_nexus()
    logger.debug('reading metadata from Vicon Nexus')
    subjname = get_subjectnames()
    params_available = vicon.GetSubjectParamNames(subjname)
    subj_params = defaultdict(lambda: None)
    subj_params.update(
        {
            par: _get_nexus_subject_param(vicon, subjname, par)
            for par in params_available
        }
    )
    trialname = _get_trialname()
    if not trialname:
        raise GaitDataError('No trial loaded in Nexus')
    sessionpath = get_sessionpath()
    markers = _get_marker_names(vicon)
    # get foot strike and toeoffevents. GetEvents() indices seem to often be 1
    # frame less than on Nexus display - only happens with ROI?
    lstrikes = vicon.GetEvents(subjname, "Left", "Foot Strike")[0]
    rstrikes = vicon.GetEvents(subjname, "Right", "Foot Strike")[0]
    ltoeoffs = vicon.GetEvents(subjname, "Left", "Foot Off")[0]
    rtoeoffs = vicon.GetEvents(subjname, "Right", "Foot Off")[0]
    events = TrialEvents(
        rstrikes=rstrikes, lstrikes=lstrikes, rtoeoffs=rtoeoffs, ltoeoffs=ltoeoffs
    )

    # offset will be subtracted from event frame numbers to get correct
    # 0-based index for frame data. for Nexus, it is always 1 (Nexus uses
    # 1-based frame numbering)
    offset = 1
    length = vicon.GetFrameCount()
    framerate = vicon.GetFrameRate()
    # get analog rate. this may not be mandatory if analog devices
    # are not used, but currently it needs to succeed.
    devids = vicon.GetDeviceIDs()
    if not devids:
        raise GaitDataError('Cannot determine analog rate')
    else:
        # rates may be 0 for some devices, we just pick the maximum as "the rate"
        analogrates = [vicon.GetDeviceDetails(id)[2] for id in devids]
        analograte = max(rate for rate in analogrates if _isfloat(rate))
    if analograte == 0.0:
        raise GaitDataError('Cannot determine analog rate')
    samplesperframe = analograte / framerate
    logger.debug(
        'offset @ %d, %d frames, framerate %d Hz, %d samples per '
        'frame' % (offset, length, framerate, samplesperframe)
    )
    # get n of forceplates
    fp_devids = [
        id_ for id_ in devids if vicon.GetDeviceDetails(id_)[1].lower() == 'forceplate'
    ]

    return {
        'trialname': trialname,
        'sessionpath': sessionpath,
        'offset': offset,
        'framerate': framerate,
        'analograte': analograte,
        'name': subjname,
        'subj_params': subj_params,
        'events': events,
        'length': length,
        'samplesperframe': samplesperframe,
        'n_forceplates': len(fp_devids),
        'markers': markers,
    }


def _get_emg_data(vicon):
    """Read EMG data from Nexus. Uses the configured EMG device name."""
    return _get_analog_data(vicon, cfg.emg.devname)


def _get_accelerometer_data(vicon):
    """Read accelerometer data from Nexus. Uses the configured acc device name."""
    return _get_analog_data(vicon, cfg.analog.accelerometer_devname)


def _get_analog_data(vicon, devname):
    """Read analog data from Vicon Nexus.

    Parameters
    ----------
    vicon : ViconNexus
        The SDK object.
    devname : str
        The analog device name, set in Nexus configuration. E.g. 'Myon EMG'.

    Returns
    -------
    dict
        Dict with keys 't' (time points corresponding to data samples) and
        'data' (the analog data as shape (N,) ndarray, for each output channel).
    """
    # match devname exactly (not case-sensitive though)
    ids = [
        id_
        for id_ in vicon.GetDeviceIDs()
        if vicon.GetDeviceDetails(id_)[0].lower() == devname.lower()
    ]
    if len(ids) > 1:
        raise GaitDataError('Multiple matching analog devices for %s' % devname)
    elif len(ids) == 0:
        raise GaitDataError('No matching analog devices for %s' % devname)
    dev_id = ids[0]
    dname, dtype, drate, outputids, _, _ = vicon.GetDeviceDetails(dev_id)
    # gather device outputs; there does not seem to be any reliable way to
    # identify output IDs that have actual EMG signal, so we use the heuristic
    # of units being volts. this may lead to inclusion of some channels (e.g.
    # Foot Switch on Noraxon) that are not actually EMG
    emg_outputids = [
        outputid
        for outputid in outputids
        if vicon.GetDeviceOutputDetails(dev_id, outputid)[2] == 'volt'
    ]

    data = dict()
    for outputid in emg_outputids:
        # get list of channel names and IDs
        outputname, _, _, _, chnames, chids = vicon.GetDeviceOutputDetails(
            dev_id, outputid
        )
        for chid in chids:
            chdata, _, chrate = vicon.GetDeviceChannel(dev_id, outputid, chid)
            chname = chnames[chid - 1]  # chids start from 1
            # in case of multiple output ids (e.g. Noraxon), the channel
            # names may not be unique, so try to generate unique names by
            # merging output name and channel name
            if len(emg_outputids) > 1:
                logger.warning(
                    'merging output %s and channel name %s for a unique name'
                    % (outputname, chname)
                )
                chname = '%s_%s' % (outputname, chname)
            if chname in data:
                raise RuntimeError('duplicate EMG channel; check Nexus device settings')
            data[chname] = np.array(chdata)
    # WIP: sanity checks for data (channel lengths equal, etc.)
    t = np.arange(len(chdata)) / drate  # time axis
    return {'t': t, 'data': data}


def _get_forceplate_ids(vicon):
    """Return Nexus forceplate device IDs"""
    return [
        id_
        for id_ in vicon.GetDeviceIDs()
        if vicon.GetDeviceDetails(id_)[1].lower() == 'forceplate'
    ]


def set_forceplate_data(vicon, fp_index, data, kind='Force'):
    """Set forceplate data in Nexus.

    This always sets the data in the device local frame. To set data in the
    global frame, you need to get the local->global transformation from Nexus,
    invert it, and apply the resulting global->local transformation to inputs.

    Parameters
    ----------
    vicon : ViconNexus
        The SDK object.
    fp_index : int
        The index of the forceplate (0...N). Note that this is not the same as
        Nexus device ID.
    data : ndarray
        Tx3 array of data, where T is number of analog frames in current data. T
        needs to equal the number of analog samples in the current Nexus trial
        (not the ROI, but whole trial).
    kind : str, optional
        Kind of data to write. Can be 'Force', 'Moment' or 'CoP', by default
        'Force'.
    """
    kinds = ['Force', 'Moment', 'CoP']
    if kind not in kinds:
        raise ValueError('kind argument needs to be one of %s' % ', '.join(kinds))
    fpids = _get_forceplate_ids(vicon)
    if not fpids:
        raise RuntimeError('no forceplates detected')
    else:
        try:
            fpid = fpids[fp_index]
        except IndexError:
            raise RuntimeError(
                'Invalid plate index %d (detected %d forceplates)'
                % (fp_index, len(fpids))
            )
    outputid = vicon.GetDeviceOutputIDFromName(fpid, kind)
    for dim, data_dim in zip('xyz', data.T):
        chname = kind[0] + dim  # e.g. 'Fx'
        chid = vicon.GetDeviceChannelIDFromName(fpid, outputid, chname)
        vicon.SetDeviceChannel(fpid, outputid, chid, data_dim)


def _get_1_forceplate_data(vicon, devid, coords='global'):
    """Read data of a single forceplate from Nexus.

    Parameters
    ----------
    vicon : ViconNexus
        The SDK object.
    devid : int
        The device id.
    coords : str, optional
        Whether to return data in 'global' or 'local' coordinate system.

    Returns
    -------
    dict
        Force dict with the following keys and values:

        F : ndarray
            Nx3 matrix of force.
        Ftot : ndarray
            Nx1 matrix of force magnitude.
        M : ndarray
            Nx3 matrix of moment.
        CoP : ndarray
            Nx3 matrix of center of pressure.
        wR : ndarray
            3x3 rotation matrix for local -> global frame.
        wT : ndarray
            1x3 transformation vector for local -> global frame.
        plate_corners : ndarray
            4x3 matrix of the four plate corners.

    All data is returned in the coordinate frame specified by the coords
    parameter. wR and wT are always returned as local -> world transformation.
    """
    if coords == 'global':
        getter_fun = vicon.GetDeviceChannelGlobal
    elif coords == 'local':
        getter_fun = vicon.GetDeviceChannel
    else:
        raise ValueError('Invalid coords argument, must be "global" or "local"')
    logger.debug('reading forceplate data from devid %d' % devid)
    dname, dtype, drate, outputids, nfp, _ = vicon.GetDeviceDetails(devid)
    kinds = ['Force', 'Moment', 'CoP']
    alldata = dict()
    for kind in kinds:
        outputid = vicon.GetDeviceOutputIDFromName(devid, kind)
        datalist = list()
        for dim in 'xyz':
            chname = kind[0] + dim  # e.g. 'Fx'
            chid = vicon.GetDeviceChannelIDFromName(devid, outputid, chname)
            data, chready, chrate = getter_fun(devid, outputid, chid)
            datalist.append(data)
        alldata[kind] = np.array(datalist).T
    Ftot = np.linalg.norm(alldata['Force'], axis=1)
    cop = alldata['CoP']
    # translation and rotation matrices from local to global coordinates
    wR = np.array(nfp.WorldR).reshape(3, 3)
    wT = np.array(nfp.WorldT)
    # get plate coords, convert to global frame if needed
    cor = np.stack([nfp.LowerBounds, nfp.UpperBounds])
    if coords == 'global':
        cor = _change_coords(cor, wR, wT)
    plate_corners = np.array(
        [
            cor[0, :],
            [cor[0, 0], cor[1, 1], cor[0, 2]],
            cor[1, :],
            [cor[1, 0], cor[0, 1], cor[0, 2]],
        ]
    )
    lb = np.min(cor, axis=0)
    ub = np.max(cor, axis=0)
    # check that CoP stays inside plate boundaries
    cop_ok = np.logical_and(cop[:, 0] >= lb[0], cop[:, 0] <= ub[0]).all()
    cop_ok &= np.logical_and(cop[:, 1] >= lb[1], cop[:, 1] <= ub[1]).all()
    if not cop_ok:
        logger.warning('center of pressure outside plate boundaries, clipping to plate')
        cop[:, 0] = np.clip(cop[:, 0], lb[0], ub[0])
        cop[:, 1] = np.clip(cop[:, 1], lb[1], ub[1])
    return {
        'F': alldata['Force'],
        'M': alldata['Moment'],
        'Ftot': Ftot,
        'CoP': cop,
        'wR': wR,
        'wT': wT,
        'plate_corners': plate_corners,
    }


def _get_forceplate_data(vicon):
    """Read data of all forceplates from Nexus.

    See read_data.get_forceplate_data() for details.
    """
    # get forceplate ids
    logger.debug('reading forceplate data from Vicon Nexus')
    devids = _get_forceplate_ids(vicon)
    if not devids:
        logger.debug('no forceplates detected')
        return None
    logger.debug('detected %d forceplate(s)' % len(devids))
    # filter by device names, if they are configured
    if cfg.autoproc.nexus_forceplate_devnames:
        devids = [
            id
            for id in devids
            if vicon.GetDeviceDetails(id)[0] in cfg.autoproc.nexus_forceplate_devnames
        ]
    return [_get_1_forceplate_data(vicon, devid) for devid in devids]


def _swap_markers(vicon, marker1, marker2):
    """Swap trajectories of given two markers in the current trial"""
    subj = get_subjectnames()
    m1 = vicon.GetTrajectory(subj, marker1)
    m2 = vicon.GetTrajectory(subj, marker2)
    vicon.SetTrajectory(subj, marker2, m1[0], m1[1], m1[2], m1[3])
    vicon.SetTrajectory(subj, marker1, m2[0], m2[1], m2[2], m2[3])


def _get_marker_data(vicon, markers, ignore_missing=False):
    """Get position data for specified markers.

    See read_data.get_marker_data for details.
    """
    if not isinstance(markers, list):
        markers = [markers]
    subj = get_subjectnames()
    mkrdata = dict()
    for marker in markers:
        x, y, z, _ = vicon.GetTrajectory(subj, marker)
        if len(x) == 0:
            if ignore_missing:
                logger.warning('Cannot read trajectory %s from Nexus' % marker)
                continue
            else:
                raise GaitDataError(
                    'Cannot read marker trajectory from Nexus: %s' % marker
                )
        mkrdata[marker] = np.array([x, y, z]).transpose()
    return mkrdata


def _get_model_data(vicon, model):
    """Read model output variables (e.g. Plug-in Gait).

    See read_data.get_model_data for details.
    """
    modeldata = dict()
    var_dims = (3, vicon.GetFrameCount())
    subj = get_subjectnames()
    for var in model.read_vars:
        nums, bools = vicon.GetModelOutput(subj, var)
        if nums:
            data = np.squeeze(np.array(nums))
        else:
            logger.info('cannot read variable %s, returning nans' % var)
            data = np.empty(var_dims)
            data[:] = np.nan
        modeldata[var] = data
    return modeldata


def _create_events(vicon, context, strikes, toeoffs):
    """Create foot strike and toeoff events in Nexus"""
    logger.debug('marking events in Nexus')
    side_str = 'Right' if context == 'R' else 'Left'
    subjectname = get_subjectnames()
    for fr in strikes:
        vicon.CreateAnEvent(subjectname, side_str, 'Foot Strike', int(fr + 1), 0)
    for fr in toeoffs:
        vicon.CreateAnEvent(subjectname, side_str, 'Foot Off', int(fr + 1), 0)


def rigid_body_extrapolate(
    vicon,
    ref_trial,
    extrap_trials,
    ref_markers,
    extrap_markers,
    ref_frame=None,
    save_trials=True,
):
    """Extrapolate missing marker positions from one trial to others.

    Rigid body extrapolation: all markers are assumed to be part of a rigid
    cluster. The reference markers (ref_markers) must be present in all trials.
    The markers to be extrapolated (extrap_markers) must be present in the
    reference trial. The relative position of the reference marker cluster is
    computed in the extrapolation trials, and the positions of the missing
    markers are extrapolated based on that.

    Note that the reference trial and extrapolation trial can be the same. In
    that case, set ref_frame to a frame where all the markers are present. The
    other frames will then be extrapolated.

    The extrapolated marker positions are written back into Nexus and
    vicon.SaveTrial() is called to save each trial.

    Parameters
    ----------
    vicon : ViconNexus
        A ViconNexus SDK object.
    ref_trial : str
        Filename of the reference trial. Give the filename without the
        'Trial.enf' extension, e.g. 'my_session\my_trial01'. All markers must be
        present in the reference trial, at least at the frame given by
        ref_frame.
    extrap_trials : list
        The trials on which the extrapolation should be performed.
    ref_markers : list
        The reference markers to extrapolate from.
    extrap_markers : list
        The markers to extrapolate.
    ref_frame : int or None
        The reference frame to use from the reference trial. If not given, defaults to 0.
    save_trials : bool
        Whether to save the extrapolated trials.
    """
    if ref_frame is None:
        ref_frame = 0
    if not isinstance(extrap_trials, list):
        extrap_trials = [extrap_trials]
    # read data from reference trial
    _open_trial(ref_trial)
    allmarkers = ref_markers + extrap_markers
    try:
        mdata_ref = _get_marker_data(vicon, allmarkers)
    except GaitDataError:
        raise GaitDataError('cannot read markers from reference trial %s' % ref_trial)
    mdata_ref_ = np.row_stack(
        [mdata_ref[marker][ref_frame, :] for marker in allmarkers]
    )
    # read reference marker data from extrapolation trials
    for extrap_trial in extrap_trials:
        extrap_coords = defaultdict(list)
        _open_trial(extrap_trial)
        subjname = get_subjectnames()
        try:
            mdata_ref = _get_marker_data(vicon, ref_markers)
        except GaitDataError:
            raise GaitDataError(
                'cannot read markers from extrapolation trial %s' % extrap_trial
            )
        for frame in range(vicon.GetFrameCount()):
            mdata_thisframe = np.row_stack(
                [mdata_ref[marker][frame, :] for marker in ref_markers]
            )
            mdata_extrap = _rigid_body_extrapolate_markers(mdata_ref_, mdata_thisframe)
            for marker, pos in zip(extrap_markers, mdata_extrap):
                extrap_coords[marker].append(pos)
        # write extrapolated data
        for marker, vals in extrap_coords.items():
            vals_x, vals_y, vals_z = np.array(vals).T
            data_exists = [True] * vicon.GetFrameCount()
            vicon.SetTrajectory(
                subjname,
                marker,
                vals_x,
                vals_y,
                vals_z,
                data_exists,
            )
        if save_trials:
            vicon.SaveTrial(cfg.autoproc.nexus_timeout)
